#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : timer.py
# Author            : Jiahao Yao
# Email             : jiahaoyao.math@gmail.com
# Date              : 04.04.2020
# Last Modified Date: 04.12.2020
# Last Modified By  : Jiahao Yao
#
# This file is part of the Flow codebase
# Distributed under MIT license

from collections import defaultdict
import time

class TimeIt(object):
    """Time the module
    # https://github.com/gkahn13/badgr/blob/92881c5df057aa50acbb92160afb51bea24d629f/src/badgr/utils/python_utils.py#L175-L228
    Example:
    with timeit('sth'):
        dosomething()

    for line in str(timeit).split('\n'):
        print(line)
    """
    def __init__(self, prefix=''):
        self.prefix = prefix
        self.start_times = dict()
        self.elapsed_times = defaultdict(int)

        self._with_name_stack = []

    def __call__(self, name):
        self._with_name_stack.append(name)
        return self

    def __enter__(self):
        self.start(self._with_name_stack[-1])
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        timeit.stop(self._with_name_stack.pop())

    def start(self, name):
        assert(name not in self.start_times)
        self.start_times[name] = time.time()

    def stop(self, name):
        assert(name in self.start_times)
        self.elapsed_times[name] += time.time() - self.start_times[name]
        self.start_times.pop(name)

    def elapsed(self, name):
        return self.elapsed_times[name]

    def display(self):
    	for line in str(self).split('\n'):
    	    print(line)

    def reset(self):
        self.start_times = dict()
        self.elapsed_times = defaultdict(int)

    def __str__(self):
        s = ''
        names_elapsed = sorted(self.elapsed_times.items(), key=lambda x: x[1], reverse=True)
        for name, elapsed in names_elapsed:
            if 'total' not in self.elapsed_times:
                s += '{0}: {1: <10} {2:.1f}\n'.format(self.prefix, name, elapsed)
            else:
                assert(self.elapsed_times['total'] >= max(self.elapsed_times.values()))
                pct = 100. * elapsed / self.elapsed_times['total']
                s += '{0}: {1: <10} {2:.1f} ({3:.1f}%)\n'.format(self.prefix, name, elapsed, pct)
        if 'total' in self.elapsed_times:
            times_summed = sum([t for k, t in self.elapsed_times.items() if k != 'total'])
            other_time = self.elapsed_times['total'] - times_summed
            assert(other_time >= 0)
            pct = 100. * other_time / self.elapsed_times['total']
            s += '{0}: {1: <10} {2:.1f} ({3:.1f}%)\n'.format(self.prefix, 'other', other_time, pct)
        return s

timeit = TimeIt()


class Timer:    
    def __enter__(self):
        self.start = time.perf_counter()
        return self

    def __exit__(self, *args):
        self.end = time.perf_counter()
        self.interval = self.end - self.start



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : cache.py
# Author            : Chi Han, Jiayuan Mao
# Email             : haanchi@gmail.com, maojiayuan@gmail.com
# Date              : 06.08.2019
# Last Modified Date: 30.08.2019
# Last Modified By  : Chi Han, Jiayuan Mao
#
# This file is part of the VCML codebase
# Distributed under MIT license


import os
from utility.common import load, dump, make_parent_dir


class Cache:
    def __init__(self, name, logger, args):
        self.name = name
        self.logger = logger
        self.args = args

        if self.args.clear_cache:
            logger('Clearing cache: {self.filename}')
            self.rm_file(empty_ok=True)

    def exist(self):
        output = os.path.exists(self.filename)
        return output

    def load(self):
        self.logger(f'Using cache: {self.filename}')
        output = load(self.filename)
        return output

    def cache(self, obj):
        self.logger(f'Saving cache: {self.filename}')
        make_parent_dir(self.filename)
        dump(obj, self.filename)
        self.obj = obj

    def rm_file(self, empty_ok=False):
        if empty_ok and not os.path.exists(self.filename):
            return
        os.remove(self.filename)

    @property
    def filename(self):
        output = os.path.join(
            self.args.cache_dir, self.name+'.pkl'
        )
        return output

    def __enter__(self):
        if self.exist():
            self.obj = self.load()
        else:
            self.logger(f'Cache not found: {self.filename}')
            self.obj = None
        return self

    def __exit__(self, _type, _value, _traceback):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : math_utils.py
# Author            : Chi Han, Jiayuan Mao
# Email             : haanchi@gmail.com, maojiayuan@gmail.com
# Date              : 09.08.2019
# Last Modified Date: 09.08.2019
# Last Modified By  : Chi Han, Jiayuan Mao
#
# This file is part of the VCML codebase
# Distributed under MIT license
#
# math tools

import numpy as np
import torch
import torch.nn.functional as F

from .tensor_utils import to_numpy


def ceiling_division(x, y):
    return (x-1) // y + 1


def min_fn(*xs):
    output = xs[0]
    for x in xs[1:]:
        output = torch.min(output, x)
    return output


def max_fn(x, y):
    return torch.max(x, y)


def equal_ratio(x, y):
    match = equal_items(x, y)
    return match.mean()


def equal_items(x, y):
    if isinstance(x, np.ndarray) or isinstance(y, np.ndarray):
        match = (np.array(x) == np.array(y)).astype(float)
    else:
        match = (x == y).float()
    return match


def recall(x, y):
    match = to_numpy(x) * to_numpy(y)
    return match.sum() / y.sum()


def arange(n):
    return list(range(n))


def logit_and(x, y):
    max_ = max_fn(x, y)
    min_ = min_fn(x, y)
    residue_true = -min_fn(max_, max_ - min_)

    return min_ - F.softplus(residue_true)


def logit_exist(x, y):
    return x + F.softplus(-y)


def log_and(x, y):
    return log(x) + log(y)


def log_or(x, y):
    return -logit_and(-x, -y) + log_and(-x, -y)
    # return log(max_fn(x, y))


def log(x):
    return -F.softplus(-x)


def log_xor(x, y):
    return F.softplus(y - x) - F.softplus(-x) - F.softplus(y)


def log_xand(x, y):
    return log_xor(-x, y)


def logit_xor(x, y):
    return x + F.softplus(y - x) - F.softplus(y + x)


def logit_xand(x, y):
    return -logit_xor(x, y)



import io
import re
import tarfile
import xml.etree.ElementTree
from urllib.error import HTTPError

import arxiv2bib
import bibtexparser
import requests
from pybtex.database import parse_string
from pybtex.utils import OrderedCaseInsensitiveDict
from requests.exceptions import RequestException

"""
This file contains various utility functions.
"""
import argparse
import re
import unicodedata
from itertools import chain, islice


def parse_args():
    desc = "get the bib for an arxiv paper"
    parser = argparse.ArgumentParser(description=desc)
    arg_lists = []

    def add_argument_group(name):
        arg = parser.add_argument_group(name)
        arg_lists.append(arg)
        return arg

    # arxiv arg
    arx_arg = add_argument_group("arxiv")

    parser.add_argument("--input", "-i", type=argparse.FileType("r"), help="input file")
    parser.add_argument("--output", "-o", help="bibtex output file")

    args = parser.parse_known_args()

    return args


# Huge URL regex taken from https://gist.github.com/gruber/8891611
URL_REGEX = re.compile(
    r"(?i)\b((?:https?:(?:/{1,3}|[a-z0-9%])|[a-z0-9.\-]+[.](?:com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw)/)(?:[^\s()<>{}\[\]]+|\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\))+(?:\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’])|(?:(?<!@)[a-z0-9]+(?:[.\-][a-z0-9]+)*[.](?:com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw)\b/?(?!@)))"
)
_SLUGIFY_STRIP_RE = re.compile(r"[^\w\s-]")
_SLUGIFY_HYPHENATE_RE = re.compile(r"[\s]+")


def replace_all(text, replace_dict):
    """
    Replace multiple strings in a text.
    .. note::
        Replacements are made successively, without any warranty on the order \
        in which they are made.
    :param text: Text to replace in.
    :param replace_dict: Dictionary mapping strings to replace with their \
            substitution.
    :returns: Text after replacements.
    >>> replace_all("foo bar foo thing", {"foo": "oof", "bar": "rab"})
    'oof rab oof thing'
    """
    for i, j in replace_dict.items():
        text = text.replace(i, j)
    return text


def map_or_apply(function, param):
    """
    Map the function on ``param``, or apply it, depending whether ``param`` \
            is a list or an item.
    :param function: The function to apply.
    :param param: The parameter to feed the function with (list or item).
    :returns: The computed value or ``None``.
    """
    try:
        if isinstance(param, list):
            return [next(iter(function(i))) for i in param]
        else:
            return next(iter(function(param)))
    except StopIteration:
        return None


def clean_whitespaces(text):
    """
    Remove multiple whitespaces from text. Also removes leading and trailing \
    whitespaces.
    :param text: Text to remove multiple whitespaces from.
    :returns: A cleaned text.
    >>> clean_whitespaces("this  is    a text with    spaces")
    'this is a text with spaces'
    """
    return " ".join(text.strip().split())


def remove_duplicates(some_list):
    """
    Remove the duplicates from a list.
    :param some_list: List to remove duplicates from.
    :returns: A list without duplicates.
    >>> remove_duplicates([1, 2, 3, 1])
    [1, 2, 3]
    >>> remove_duplicates([1, 2, 1, 2])
    [1, 2]
    """
    return list(set(some_list))


def batch(iterable, size):
    """
    Get items from a sequence a batch at a time.
    .. note:
        Adapted from
        https://code.activestate.com/recipes/303279-getting-items-in-batches/.
    .. note:
        All batches must be exhausted immediately.
    :params iterable: An iterable to get batches from.
    :params size: Size of the batches.
    :returns: A new batch of the given size at each time.
    >>> [list(i) for i in batch([1, 2, 3, 4, 5], 2)]
    [[1, 2], [3, 4], [5]]
    """
    item = iter(iterable)
    while True:
        batch_iterator = islice(item, size)
        try:
            yield chain([next(batch_iterator)], batch_iterator)
        except StopIteration:
            return


def remove_urls(text):
    """
    Remove URLs from a given text (only removes http, https and naked domains \
    URLs).
    :param text: The text to remove URLs from.
    :returns: The text without URLs.
    >>> remove_urls("foobar http://example.com https://example.com foobar")
    'foobar foobar'
    """
    return clean_whitespaces(URL_REGEX.sub("", text))


def slugify(value):
    """
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens to have nice filenames.
    From Django's "django/template/defaultfilters.py".
    >>> slugify("El pingüino Wenceslao hizo kilómetros bajo exhaustiva lluvia y frío, añoraba a su querido cachorro. ortez ce vieux whisky au juge blond qui fume sur son île intérieure, à Γαζέες καὶ μυρτιὲς δὲν θὰ βρῶ πιὰ στὸ χρυσαφὶ ξέφωτο いろはにほへとちりぬるを Pchnąć w tę łódź jeża lub ośm skrzyń fig กว่าบรรดาฝูงสัตว์เดรัจฉาน")
    'El_pinguino_Wenceslao_hizo_kilometros_bajo_exhaustiva_lluvia_y_frio_anoraba_a_su_querido_cachorro_ortez_ce_vieux_whisky_au_juge_blond_qui_fume_sur_son_ile_interieure_a_Pchnac_w_te_odz_jeza_lub_osm_skrzyn_fig'
    """
    try:
        unicode_type = unicode
    except NameError:
        unicode_type = str
    if not isinstance(value, unicode_type):
        value = unicode_type(value)
    value = (
        unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    )
    value = unicode_type(_SLUGIFY_STRIP_RE.sub("", value).strip())
    return _SLUGIFY_HYPHENATE_RE.sub("_", value)


# https://github.com/Phyks/libbmc/blob/master/libbmc/repositories/arxiv.py

# Append arXiv to the valid identifiers list
__valid_identifiers__ = ["repositories.arxiv"]


ARXIV_IDENTIFIER_FROM_2007 = r"\d{4}\.\d{4,5}(v\d+)?"
ARXIV_IDENTIFIER_BEFORE_2007 = (
    r"("
    + (
        "|".join(
            [
                "astro-ph.GA",
                "astro-ph.CO",
                "astro-ph.EP",
                "astro-ph.HE",
                "astro-ph.IM",
                "astro-ph.SR",
                "cond-math.dis-nn",
                "cond-math.mtrl-sci",
                "cond-math.mes-hall",
                "cond-math.other",
                "cond-math.quant-gas",
                "cond-math.soft",
                "cond-math.stat-mech",
                "cond-math.str-el",
                "cond-math.supr-con",
                "gr-qc",
                "hep-ex",
                "hep-lat",
                "hep-ph",
                "hep-th",
                "math-ph",
                "nlin.AO",
                "nlin.CG",
                "nlin.CD",
                "nlin.SI",
                "nlin.PS",
                "nucl-ex",
                "nucl-th",
                "physics.acc-ph",
                "physics.ao-ph",
                "physics.atom-ph",
                "physics.atm-clus",
                "physics.bio-ph",
                "physics.chem-ph",
                "physics.class-ph",
                "physics.comp-ph",
                "physics.data-an",
                "physics.flu-dyn",
                "physics.gen-ph",
                "physics.geo-ph",
                "physics.hist-ph",
                "physics.ins-det",
                "physics.med-ph",
                "physics.optics",
                "physics.ed-ph",
                "physics.soc-ph",
                "physics.plasm-ph",
                "physics.pop-ph",
                "physics.space-ph",
                "physics.quant-ph",
                "math.AG",
                "math.AT",
                "math.AP",
                "math.CT",
                "math.CA",
                "math.CO",
                "math.AC",
                "math.CV",
                "math.DG",
                "math.DS",
                "math.FA",
                "math.GM",
                "math.GN",
                "math.GT",
                "math.GR",
                "math.HO",
                "math.IT",
                "math.KT",
                "math.LO",
                "math.MP",
                "math.MG",
                "math.NT",
                "math.NA",
                "math.OA",
                "math.OC",
                "math.PR",
                "math.QA",
                "math.RT",
                "math.RA",
                "math.SP",
                "math.ST",
                "math.SG",
                "cs.AI",
                "cs.CL",
                "cs.CC",
                "cs.CE",
                "cs.CG",
                "cs.GT",
                "cs.CV",
                "cs.CY",
                "cs.CR",
                "cs.DS",
                "cs.DB",
                "cs.DL",
                "cs.DM",
                "cs.DC",
                "cs.ET",
                "cs.FL",
                "cs.GL",
                "cs.GR",
                "cs.AR",
                "cs.HC",
                "cs.IR",
                "cs.IT",
                "cs.LG",
                "cs.LO",
                "cs.MS",
                "cs.MA",
                "cs.MM",
                "cs.NI",
                "cs.NE",
                "cs.NA",
                "cs.OS",
                "cs.OH",
                "cs.PF",
                "cs.PL",
                "cs.RO",
                "cs.SI",
                "cs.SE",
                "cs.SD",
                "cs.SC",
                "cs.SY",
                "q-bio.BM",
                "q-bio.CB",
                "q-bio.GN",
                "q-bio.MN",
                "q-bio.NC",
                "q-bio.OT",
                "q-bio.PE",
                "q-bio.QM",
                "q-bio.SC",
                "q-bio.TO",
                "q-fin.CP",
                "q-fin.EC",
                "q-fin.GN",
                "q-fin.MF",
                "q-fin.PM",
                "q-fin.PR",
                "q-fin.RM",
                "q-fin.ST",
                "q-fin.TR",
                "stat.AP",
                "stat.CO",
                "stat.ML",
                "stat.ME",
                "stat.OT",
                "stat.TH",
            ]
        )
    )
    + r")/\d+"
)
# Regex is fully enclosed in a group for findall to match it all
REGEX = re.compile(
    "((arxiv:)?(("
    + ARXIV_IDENTIFIER_FROM_2007
    + ")|("
    + ARXIV_IDENTIFIER_BEFORE_2007
    + ")))",
    re.IGNORECASE,
)

# Base arXiv URL used as id sometimes
ARXIV_URL = "http://arxiv.org/abs/{arxiv_id}"
# Eprint URL used to download sources
ARXIV_EPRINT_URL = "http://arxiv.org/e-print/{arxiv_id}"


def get_latest_version(arxiv_id):
    """
    Find the latest version of a given arXiv eprint.
    :param arxiv_id: The (canonical) arXiv ID to query.
    :returns: The latest version on eprint as a string, or ``None``.
    >>> get_latest_version('1401.2910')
    '1401.2910v1'
    >>> get_latest_version('1401.2910v1')
    '1401.2910v1'
    >>> get_latest_version('1506.06690v1')
    '1506.06690v2'
    >>> get_latest_version('1506.06690')
    '1506.06690v2'
    """
    # Get updated bibtex
    # Trick: strip the version from the arXiv id, to query updated BibTeX for
    # the preprint and not the specific version
    arxiv_preprint_id = strip_version(arxiv_id)
    updated_bibtex = bibtexparser.loads(get_bibtex(arxiv_preprint_id))
    updated_bibtex = next(iter(updated_bibtex.entries_dict.values()))

    try:
        return updated_bibtex["eprint"]
    except KeyError:
        return None


def strip_version(arxiv_id):
    """
    Remove the version suffix from an arXiv id.
    :param arxiv_id: The (canonical) arXiv ID to strip.
    :returns: The arXiv ID without the suffix version
    >>> strip_version('1506.06690v1')
    '1506.06690'
    >>> strip_version('1506.06690')
    '1506.06690'
    """
    return re.sub(r"v\d+\Z", "", arxiv_id)


def is_valid(arxiv_id):
    """
    Check that a given arXiv ID is a valid one.
    :param arxiv_id: The arXiv ID to be checked.
    :returns: Boolean indicating whether the arXiv ID is valid or not.
    >>> is_valid('1506.06690')
    True
    >>> is_valid('1506.06690v1')
    True
    >>> is_valid('arXiv:1506.06690')
    True
    >>> is_valid('arXiv:1506.06690v1')
    True
    >>> is_valid('arxiv:1506.06690')
    True
    >>> is_valid('arxiv:1506.06690v1')
    True
    >>> is_valid('math.GT/0309136')
    True
    >>> is_valid('abcdf')
    False
    >>> is_valid('bar1506.06690foo')
    False
    >>> is_valid('mare.GG/0309136')
    False
    """
    match = REGEX.match(arxiv_id)
    return (match is not None) and (match.group(0) == arxiv_id)


def get_bibtex(arxiv_id):
    """
    Get a BibTeX entry for a given arXiv ID.
    .. note::
        Using awesome https://pypi.python.org/pypi/arxiv2bib/ module.
    :param arxiv_id: The canonical arXiv id to get BibTeX from.
    :returns: A BibTeX string or ``None``.
    >>> get_bibtex('1506.06690')
    "@article{1506.06690v2,\\nAuthor        = {Lucas Verney and Lev Pitaevskii and Sandro Stringari},\\nTitle         = {Hybridization of first and second sound in a weakly-interacting Bose gas},\\nEprint        = {1506.06690v2},\\nDOI           = {10.1209/0295-5075/111/40005},\\nArchivePrefix = {arXiv},\\nPrimaryClass  = {cond-mat.quant-gas},\\nAbstract      = {Using Landau's theory of two-fluid hydrodynamics we investigate the sound\\nmodes propagating in a uniform weakly-interacting superfluid Bose gas for\\nvalues of temperature, up to the critical point. In order to evaluate the\\nrelevant thermodynamic functions needed to solve the hydrodynamic equations,\\nincluding the temperature dependence of the superfluid density, we use\\nBogoliubov theory at low temperatures and the results of a perturbative\\napproach based on Beliaev diagrammatic technique at higher temperatures.\\nSpecial focus is given on the hybridization phenomenon between first and second\\nsound which occurs at low temperatures of the order of the interaction energy\\nand we discuss explicitly the behavior of the two sound velocities near the\\nhybridization point.},\\nYear          = {2015},\\nMonth         = {Jun},\\nUrl           = {http://arxiv.org/abs/1506.06690v2},\\nFile          = {1506.06690v2.pdf}\\n}"
    >>> get_bibtex('1506.06690v1')
    "@article{1506.06690v1,\\nAuthor        = {Lucas Verney and Lev Pitaevskii and Sandro Stringari},\\nTitle         = {Hybridization of first and second sound in a weakly-interacting Bose gas},\\nEprint        = {1506.06690v1},\\nDOI           = {10.1209/0295-5075/111/40005},\\nArchivePrefix = {arXiv},\\nPrimaryClass  = {cond-mat.quant-gas},\\nAbstract      = {Using Landau's theory of two-fluid hydrodynamics we investigate the sound\\nmodes propagating in a uniform weakly-interacting superfluid Bose gas for\\nvalues of temperature, up to the critical point. In order to evaluate the\\nrelevant thermodynamic functions needed to solve the hydrodynamic equations,\\nincluding the temperature dependence of the superfluid density, we use\\nBogoliubov theory at low temperatures and the results of a perturbative\\napproach based on Beliaev diagrammatic technique at higher temperatures.\\nSpecial focus is given on the hybridization phenomenon between first and second\\nsound which occurs at low temperatures of the order of the interaction energy\\nand we discuss explicitly the behavior of the two sound velocities near the\\nhybridization point.},\\nYear          = {2015},\\nMonth         = {Jun},\\nUrl           = {http://arxiv.org/abs/1506.06690v1},\\nFile          = {1506.06690v1.pdf}\\n}"
    """
    # Fetch bibtex using arxiv2bib module
    try:
        bibtex = arxiv2bib.arxiv2bib([arxiv_id])
    except HTTPError:
        bibtex = []

    for bib in bibtex:
        if isinstance(bib, arxiv2bib.ReferenceErrorInfo):
            continue
        else:
            # Return fetched bibtex
            return bib.bibtex()
    # An error occurred, return None
    return None


def extract_from_text(text):
    """
    Extract arXiv IDs from a text.
    :param text: The text to extract arXiv IDs from.
    :returns: A list of matching arXiv IDs, in canonical form.
    >>> sorted(extract_from_text('1506.06690 1506.06690v1 arXiv:1506.06690 arXiv:1506.06690v1 arxiv:1506.06690 arxiv:1506.06690v1 math.GT/0309136 abcdf bar1506.06690foo mare.GG/0309136'))
    ['1506.06690', '1506.06690v1', 'math.GT/0309136']
    """
    # Remove the leading "arxiv:".
    return remove_duplicates(
        [
            re.sub("arxiv:", "", i[0], flags=re.IGNORECASE)
            for i in REGEX.findall(text)
            if i[0] != ""
        ]
    )


def to_url(arxiv_ids):
    """
    Convert a list of canonical DOIs to a list of DOIs URLs.
    :param dois: List of canonical DOIs.
    :returns: A list of DOIs URLs.
    >>> to_url('1506.06690')
    'http://arxiv.org/abs/1506.06690'
    >>> to_url('1506.06690v1')
    'http://arxiv.org/abs/1506.06690v1'
    """
    if isinstance(arxiv_ids, list):
        return [ARXIV_URL.format(arxiv_id=arxiv_id) for arxiv_id in arxiv_ids]
    else:
        return ARXIV_URL.format(arxiv_id=arxiv_ids)


def to_canonical(urls):
    """
    Convert a list of arXiv IDs to a list of canonical IDs.
    :param dois: A list of DOIs URLs.
    :returns: List of canonical DOIs. ``None`` if an error occurred.
    >>> to_canonical('http://arxiv.org/abs/1506.06690')
    '1506.06690'
    >>> to_canonical('http://arxiv.org/abs/1506.06690v1')
    '1506.06690v1'
    >>> to_canonical(['http://arxiv.org/abs/1506.06690'])
    ['1506.06690']
    >>> to_canonical('aaa') is None
    True
    """
    return map_or_apply(extract_from_text, urls)


def from_doi(doi):
    """
    Get the arXiv eprint id for a given DOI.
    .. note::
        Uses arXiv API. Will not return anything if arXiv is not aware of the
        associated DOI.
    :param doi: The DOI of the resource to look for.
    :returns: The arXiv eprint id, or ``None`` if not found.
    >>> from_doi('10.1209/0295-5075/111/40005')
    # Note: Test do not pass due to an arXiv API bug.
    '1506.06690'
    """
    try:
        request = requests.get(
            "http://export.arxiv.org/api/query",
            params={"search_query": "doi:%s" % (doi,), "max_results": 1},
        )
        request.raise_for_status()
    except RequestException:
        return None
    root = xml.etree.ElementTree.fromstring(request.content)
    for entry in root.iter("{http://www.w3.org/2005/Atom}entry"):
        arxiv_id = entry.find("{http://www.w3.org/2005/Atom}id").text
        # arxiv_id is an arXiv full URL. We only want the id which is the last
        # URL component.
        return arxiv_id.split("/")[-1]
    return None


def to_doi(arxiv_id):
    """
    Get the associated DOI for a given arXiv eprint.
    .. note::
        Uses arXiv API. Will not return anything if arXiv is not aware of the
        associated DOI.
    :param eprint: The arXiv eprint id.
    :returns: The DOI if any, or ``None``.
    >>> to_doi('1506.06690v1')
    '10.1209/0295-5075/111/40005'
    >>> to_doi('1506.06690')
    '10.1209/0295-5075/111/40005'
    """
    try:
        request = requests.get(
            "http://export.arxiv.org/api/query",
            params={"id_list": arxiv_id, "max_results": 1},
        )
        request.raise_for_status()
    except RequestException:
        return None
    root = xml.etree.ElementTree.fromstring(request.content)
    for entry in root.iter("{http://www.w3.org/2005/Atom}entry"):
        doi = entry.find("{http://arxiv.org/schemas/atom}doi")
        if doi is not None:
            return doi.text
    return None


def get_sources(arxiv_id):
    """
    Download sources on arXiv for a given preprint.
    .. note::
        Bulk download of sources from arXiv is not permitted by their API. \
                You should have a look at http://arxiv.org/help/bulk_data_s3.
    :param eprint: The arXiv id (e.g. ``1401.2910`` or ``1401.2910v1``) in a \
            canonical form.
    :returns: A ``TarFile`` object of the sources of the arXiv preprint or \
            ``None``.
    """
    try:
        request = requests.get(ARXIV_EPRINT_URL.format(arxiv_id=arxiv_id))
        request.raise_for_status()
        file_object = io.BytesIO(request.content)
        return tarfile.open(fileobj=file_object)
    except (RequestException, AssertionError, tarfile.TarError):
        return None


def get_bbl(arxiv_id):
    """
    Get the .bbl files (if any) of a given preprint.
    .. note::
        Bulk download of sources from arXiv is not permitted by their API. \
                You should have a look at http://arxiv.org/help/bulk_data_s3.
    :param arxiv_id: The arXiv id (e.g. ``1401.2910`` or ``1401.2910v1``) in \
            a canonical form.
    :returns: A list of the full text of the ``.bbl`` files (if any) \
            or ``None``.
    """
    tar_file = get_sources(arxiv_id)
    bbl_files = [i for i in tar_file.getmembers() if i.name.endswith(".bbl")]
    bbl_files = [
        tar_file.extractfile(member).read().decode(tarfile.ENCODING)
        for member in bbl_files
    ]
    return bbl_files


# doi2bib
# https://github.com/bibcure/doi2bib


def get_bib(doi):
    """
    Parameters
    ----------
        doi: str
    Returns
    -------
        found: bool
        bib: str
    """
    bare_url = "http://api.crossref.org/"
    url = "{}works/{}/transform/application/x-bibtex"
    url = url.format(bare_url, doi)
    r = requests.get(url)
    found = False if r.status_code != 200 else True
    bib = r.content
    bib = str(bib, "utf-8")
    return found, bib


def get_json(doi):
    """
    Parameters
    ----------
        doi: str
    Returns
    -------
        found: bool
        item: dict
            Response from crossref
    """
    bare_url = "http://api.crossref.org/"
    url = "{}works/{}"
    url = url.format(bare_url, doi)
    r = requests.get(url)
    found = False if r.status_code != 200 else True
    item = r.json()

    return found, item


def doi2_bib(doi, abbrev_journal=True, add_abstract=True):
    """
    Parameters
    ----------
        doi: str
        abbrev_journal: bool
            If True try to abbreviate the journal name
    Returns
    -------
        found: bool
        bib: str
            The bibtex string
    """
    found, bib = get_bib(doi)
    if found and abbrev_journal:

        found, item = get_json(doi)
        if found:
            abbreviated_journal = item["message"]["short-container-title"]
            if add_abstract and "abstract" in item["message"].keys():
                abstract = item["message"]["abstract"]
                bi = bibtexparser.loads(bib)
                bi.entries[0]["abstract"] = abstract
                bib = bibtexparser.dumps(bi)

            if len(abbreviated_journal) > 0:
                abbreviated_journal = abbreviated_journal[0].strip()
                bib = re.sub(
                    r"journal = \{[^>]\}",
                    "journal = {" + abbreviated_journal + "}",
                    bib,
                )
    import ipdb; ipdb.set_trace()
    return found, bib


# convert to google scholar
# https://github.com/jiamings/scholar-bibtex-keys/blob/master/gsbib/scholar_bibtex_keys.py
def obtain_replace_keys(bib_data):
    """
    Obtain Google Scholar style keys from parsed bibliography data.
    """
    keys, new_keys = [], []
    for key in bib_data.entries.keys():
        try:
            author_last_name = (
                bib_data.entries[key].persons["author"][0].last_names[0].lower()
            )
        except:
            print(key)
        try:
            year = bib_data.entries[key].fields["year"]
        except:
            year = ""
        title_first_word = (
            re.search(r"\w+", bib_data.entries[key].fields["title"]).group(0).lower()
        )
        new_key = author_last_name + year + title_first_word
        keys.append(key)
        new_keys.append(new_key)
    return keys, new_keys


def update_arxiv_information(bib_data):
    """
    Include arxiv information in journal field.
    """
    for key, entry in bib_data.entries.items():
        if "Eprint" in bib_data.entries[key].fields:
            bib_data.entries[key].fields["journal"] = "arXiv preprint arXiv:{}".format(
                bib_data.entries[key].fields["Eprint"]
            )
    return bib_data


def convert_bibtex_keys(input_file: str):
    """
    Convert keys in a bibtex file to Google Scholar format.
    @input_file: string, input file name.
    @output_file: string, output file name.
    """
    bib_data = parse_string(input_file, bib_format="bibtex")
    keys, new_keys = obtain_replace_keys(bib_data)
    new_entries = OrderedCaseInsensitiveDict()
    for key, new_key in zip(keys, new_keys):
        new_entries[new_key] = bib_data.entries[key]
    bib_data.entries = new_entries
    bib_data = update_arxiv_information(bib_data)
    return bib_data.to_string(bib_format="bibtex")


# utils
# https://github.com/bibcure/doi2bib/blob/master/doi2bib/bin/doi2bib
def save_output_bibs(bibs, output_file):
    try:
        with io.open(output_file, "w", encoding="utf-8") as bibfile:
            for bib in bibs:
                bibfile.write("{}\n".format(bib))

    except TypeError:
        print("Can't save in output file\n")
        print(bibs)


def arxb(arv_id):

    if is_valid(arv_id):
        doi_id = to_doi(arv_id)
        if doi_id:
            found, bib_file = get_bib(doi_id)
            assert found
        else:
            bib_file = get_bibtex(arv_id)
    else:
        doi_id = arv_id
        found, bib_file = get_bib(doi_id)
        assert found

    # fix the url
    return convert_bibtex_keys(bib_file).replace(r'\%2', '%2')   


def main():
    global args
    args = parse_args()

    inlinearvid = len(args[1]) > 0
    if inlinearvid:
        arvids = args[1]
    else:
        arvids = args[0].input.read()
        arvids = map(lambda a: a.strip(), arvids.split("\n"))
        arvids = filter(lambda title: title != "", arvids)

    bibs = list(map(arxb, arvids))

    if inlinearvid:
        print("\n".join(bibs))
    else:
        if args[0].output:
            save_output_bibs(bibs, args[0].output)
        else:
            outfile = args[0].input.name
            outfile = outfile.split(".")
            outfile[-2] = outfile[-2] + "_convert"
            outfile[-1] = "bib"
            outfile = ".".join(outfile)
            save_output_bibs(bibs, outfile)


if __name__ == "__main__":
    # Note down some potential useful lib: https://github.com/yuchenlin/rebiber
    # Usage I:
    # python arxb.py 1506.06690v2  2012.20323

    # Using II:
    # test.md
    # 1506.06690v2
    # 1506.06691
    # 1506.06692
    # 2012.20323
    # python arxb.py -i test.md

    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : recording.py
# Author            : Chi Han, Jiayuan Mao
# Email             : haanchi@gmail.com, maojiayuan@gmail.com
# Date              : 03.08.2019
# Last Modified Date: 27.08.2019
# Last Modified By  : Chi Han, Jiayuan Mao
#
# This file is part of the VCML codebase
# Distributed under MIT license


import os
import numpy as np
import torch
from abc import abstractmethod

from utility.common import make_parent_dir


class History:
    def __init__(self, name, capacity):
        self.capacity = capacity
        self.log = []
        self.n = 0
        self.log_interval = 1
        self.name = name

    def record(self, value):
        self.n += 1
        if self.n % self.log_interval == 0:
            self.log.append(value)
            if len(self.log) == self.capacity:
                self.log = self.log[::2]
                self.log_interval *= 2

    def plot(self, ax, label):
        axis_time = np.arange(len(self.log))
        axis_time *= self.log_interval
        axis_value = np.array(self.log)
        ax.plot(axis_time, axis_value, label=label)

    def state_dict(self):
        ckpt = {
            'n': self.n,
            'log': self.log,
            'log_interval': self.log_interval,
            'name': self.name,
            'capacity': self.capacity,
        }
        return ckpt

    def load_state_dict(self, ckpt):
        self.__dict__.update(ckpt)


class MatrixHistory:
    def __init__(self, name, dims, capacity):
        self.name = name
        self.capacity = capacity
        self.log = np.zeros(dims + (capacity,))
        self.log_indexes = np.zeros(dims[0], dtype=int)

    def record(self, row_ind, value):
        self.log[row_ind, :, self.log_indexes[row_ind] % self.capacity] = value
        self.log_indexes[row_ind] += 1

    def reset(self):
        self.log *= 0
        self.log_indexes *= 0

    def state_dict(self):
        ckpt = {
            'name': self.name,
            'capacity': self.capacity,
            'log': self.log,
            'log_indexes': self.log_indexes,
        }
        return ckpt

    def load_state_dict(self, ckpt):
        self.__dict__.update(ckpt)


class AverageRecording:
    def __init__(self, name, history_capacity):
        self.name = name
        self.current = 0

        self.history = History(
            f'{name}_history', history_capacity)
        self.value_history = History(
            f'{name}_valueHistory', history_capacity)

    @abstractmethod
    def reset(self):
        pass

    def record(self, new_value):
        assert not isinstance(new_value, torch.Tensor), 'not converted'
        new_value = new_value
        self.history.record(new_value)

        self.update_average(new_value)
        self.value_history.record(self.value)

    def __str__(self):
        return '{0}: {1:.3f}'.format(self.name, self.value)

    def state_dict(self):
        return {
            'name': self.name,
            'current': self.current,
            'history': self.history.state_dict(),
            'value_history': self.value_history.state_dict(),
        }

    def load_state_dict(self, ckpt):
        self.name = ckpt['name']
        self.current = ckpt['current']
        self.history.load_state_dict(ckpt['history'])
        self.value_history.load_state_dict(ckpt['value_history'])

    @property
    @abstractmethod
    def value(self):
        pass

    @abstractmethod
    def update_average(self):
        pass

    def plot(self, ax, label):
        self.value_history.plot(ax, label)
        ax.set_xlabel('time')
        ax.set_ylabel('value')
        ax.set_title(self.name)


class ExponentialAverage(AverageRecording):
    def __init__(self, name, capacity, momentum=0.99):
        super().__init__(name, capacity)
        self.momentum = momentum
        self.weight = 0

    def update_average(self, new_value):
        y = self.momentum
        self.current = y * self.current + (1 - y) * new_value
        self.weight = y * self.weight + (1 - y) * 1

    @property
    def value(self):
        if self.weight == 0:
            return 0
        else:
            return self.current / self.weight

    def reset(self):
        self.weight = 0
        self.current = 0

    def state_dict(self):
        ckpt = super().state_dict()
        ckpt.update({
            'weight': self.weight
        })
        return ckpt

    def load_state_dict(self, ckpt):
        super().load_state_dict(ckpt)
        self.weight = ckpt['weight']


class SimpleAverage(AverageRecording):
    def __init__(self, name, capacity):
        super().__init__(name, capacity)
        self.count = 0

    def update_average(self, new_value):
        self.current += new_value
        self.count += 1

    @property
    def value(self):
        if self.count == 0:
            return 0
        else:
            return self.current / self.count

    def reset(self):
        self.count = 0
        self.current = 0

    def state_dict(self):
        ckpt = super().state_dict()
        ckpt.update({
            'count': self.count
        })
        return ckpt

    def load_state_dict(self, ckpt):
        super().load_state_dict(ckpt)
        self.count = ckpt['count']


class CoeffMatrixRecording:
    def __init__(self, name, dim, history_capacity):
        self.name = name
        self.dim = dim
        self.history_capacity = history_capacity
        self.history = MatrixHistory(
            name,
            (dim, dim),
            history_capacity,
        )

    def record(self, row_ind, value):
        if value.ndim == 2:
            for i in range(value.shape[0]):
                self.history.record(row_ind, value[i])
        else:
            self.history.record(row_ind, value)

    def reset(self):
        self.history.reset()

    def calculate_coeff(self):
        coeff_matrix = np.zeros((self.dim, self.dim))
        log_mat = self.history.log

        for i in range(self.dim):
            x = log_mat[i, i]
            if x.var() > 0:
                for j in range(self.dim):
                    y = log_mat[i, j]
                    coeff = np.polyfit(x, y, 1)[0]
                    coeff_matrix[i, j] = coeff

        return coeff_matrix

    def state_dict(self):
        ckpt = {
            'name': self.name,
            'dim': self.dim,
            'history_capacity': self.history_capacity,
            'history': self.history.state_dict(),
        }
        return ckpt

    def load_state_dict(self, ckpt):
        self.name = ckpt['name']
        self.dim = ckpt['dim']
        self.history_capacity = ckpt['history_capacity']
        self.history.load_state_dict(ckpt['history'])


class AverageGroup:
    def __init__(self, groupname, mode, history, path):
        self.groupname = groupname
        self.mode = mode
        self.history = history
        self.path = path
        if mode == 'exponential':
            self.recording_cls = ExponentialAverage
        elif mode == 'simple':
            self.recording_cls = SimpleAverage

        self.group = {}

    def record(self, value_dict):
        for key in value_dict.keys():
            if key not in self.group:
                self.setup_new(key)

        for key, recording in self.group.items():
            if key in value_dict and value_dict[key] is not None:
                recording.record(value_dict[key])
            else:
                recording.record(recording.value)

    def setup_new(self, key):
        new_name = f'{self.groupname}_{key}'
        self.group[key] = self.recording_cls(new_name, self.history)

    def reset(self):
        for item in self.group.values():
            item.reset()

    def state_dict(self):
        ckpt = {
            'groupname': self.groupname,
            'mode': self.mode,
            'group': {key: recording.state_dict()
                      for key, recording in self.group.items()}
        }
        return ckpt

    def load_state_dict(self, ckpt):
        self.groupname = ckpt['groupname']
        self.mode = ckpt['mode']
        for key, recording in ckpt['group'].items():
            self.setup_new(key)
            self.group[key].load_state_dict(recording)

    def __str__(self):
        return ' '.join([
            '{0}: {1:.2f}'.format(key, recording.value)
            for key, recording in self.group.items()
        ])

    def __len__(self):
        return len(self.group)

    @property
    def values(self):
        return {
            key: recording.value
            for key, recording in self.group.items()
        }

    @property
    def visualize_filename(self):
        return os.path.join(
            self.path,
            'plots',
            f'{self.groupname}.jpg'
        )

    def plot_on_axes(self, key_dict, ax_array, label):
        for key, recording in self.group.items():
            if key not in key_dict:
                key_dict[key] = len(key_dict)
            ind = key_dict[key]
            recording.plot(ax_array[ind], label)

    def savefig(self, fig):
        fig.suptitle(self.groupname)
        fig.tight_layout()
        make_parent_dir(self.visualize_filename)
        fig.savefig(self.visualize_filename)



import pyperclip
import re
import argparse


def parse_args():
    desc = "replace the macro for an arxiv paper"
    parser = argparse.ArgumentParser(description=desc)
    arg_lists = []

    def add_argument_group(name):
        arg = parser.add_argument_group(name)
        arg_lists.append(arg)
        return arg

    # latex path
    arx_arg = add_argument_group("latex")
    arx_arg.add_argument("path", type=str, help="the path of the latex")
    args = parser.parse_args()

    return args


def to_raw(string):
    # some hard code
    return (
        string.replace("\\", "\\\\")
        .replace("\\\\g<", "\\g<")
        .replace("?!\\\\w", "?!\w")
    )


def translate(text, conversion_dict, before=None):
    """
    Translate words from a text using a conversion dictionary

    Arguments:
        text: the text to be translated
        conversion_dict: the conversion dictionary
        before: a function to transform the input
        (by default it will do nothing)
    """
    # if empty:
    if not text:
        return text
    # preliminary transformation:
    if before:
        t = before(text)
    else:
        t = text
    for key, value in conversion_dict.items():
        t = re.sub(key, value, t)

    return t


def latex_replace(input_file):
    assert input_file.endswith(".tex")
    output_file = input_file.replace(".tex", "_replace.tex")

    with open(input_file, "r") as f:
        input_lines = f.readlines()

    start_trans = False

    replace_dict = {}
    output_lines = []
    for line in input_lines:

        # case I:
        # \DeclareMathOperator{\Ker}{Ker}
        # ====>
        # \newcommand{\Ker}{\operatorname{Ker}}
        if line.startswith("\\DeclareMathOperator"):
            line = re.sub(
                r"\\DeclareMathOperator\**{(.*)}{(.*)}",
                r"\\newcommand{\g<1>}{\operatorname{\g<2>}}",
                line,
            )

        # case II:
        # A = '\def\hh{\hat{h}}'
        # re.sub(r'\\def(.*?){(.*)}',r'\\newcommand{\g<1>}{\g<2>}', A)
        if line.startswith("\\def"):
            line = re.sub(r"\\def(.*?){(.*)}", r"\\newcommand{\g<1>}{\g<2>}", line)

        if (
            line.startswith("\\newcommand")
            or line.startswith("\\renewcommand")
            or line.startswith("\\providecommand")
        ):
            keyword = re.findall("((new|renew|provide)command)", line)
            keyword = keyword[0][0]

            if len(re.findall(r"\[(\d)\]", line)) != 0:
                num_params = int(re.findall(r"\[(\d)\]", line)[0])
            else:
                num_params = 0

            if num_params:
                # example
                # \newcommand{\ip}[1] {\langle #1 \rangle }
                # r'\\ip{(.*)}': r'\\langle \g<1> \\rangle'

                try: 
                    line_map = re.findall(
                        r"\\{}".format(keyword) + r"\**{(.*?)}\s*\[\d\]\s*{(.*)}", line
                    )[0]
                except: 
                    line_map = re.findall(
                        r"\\{}".format(keyword) + r"\**(.*?)\s*\[\d\]\s*{(.*)}", line
                    )[0]

                assert len(line_map) == 2
                key, value = line_map

                key = key + "{(.*?)}" * num_params

                for i in range(num_params):
                    i += 1
                    value = value.replace("#{}".format(i), "\g<{}>".format(i))

                # example (nested newcommand)
                # \newcommand{\cA}{\mathcal{A}}
                # \newcommand{\das}{\Delta_{\cA}^{\cS}}

                value = translate(value, replace_dict)
                replace_dict[to_raw(key)] = to_raw(value)

            else:
                # example
                # \newcommand{\minimize}{\mathop{\textrm{minimize}}}
                # line_map = re.findall(r'\\newcommand{(.*?)}{(.*)}', line)[0]

                # star: for \renewcommand*{\thepa}{\alph{pa}}
                try: 
                    line_map = re.findall(
                        r"\\{}".format(keyword) + r"\**{(.*?)}{(.*)}", line
                    )[0]
                except: 
                    line_map = re.findall(
                        r"\\{}".format(keyword) + r"\**(.*?){(.*)}", line
                    )[0]

                assert len(line_map) == 2

                key, value = line_map
                # avoid
                # \newcommand{\be}{\mathbf{e}}
                # changes \begin
                # look ahead negative
                key = key + "(?![a-z])"
                value = translate(value, replace_dict)

                # extra space
                value = " " + value

                replace_dict[to_raw(key)] = to_raw(value)

        else:
            if start_trans:
                if len(line.strip()):
                    line = translate(line, replace_dict)

            output_lines.append(line)

        if r"\begin{document}" in line:
            start_trans = True

            print("Replace mapping: ")
            for a, b in replace_dict.items():
                print(a, ":", b)

    with open(output_file, "w") as f:
        f.write("".join(output_lines))


def main():
    global args
    args = parse_args()

    latex_replace(args.path)


if __name__ == "__main__":
    # usage:
    # 1. flap the latex: flap my/project/main.tex my/output
    # 2. replace the latex

    # testing 
    import urllib.request
    import os 

    print('Download paper')
    url = 'https://arxiv.org/e-print/2007.02151'
    fname = '2007.02151.tar.gz'
    urllib.request.urlretrieve(url, fname)

    print('Unzip the files')
    import tarfile

    tar = tarfile.open(fname, "r:gz")
    tar.extractall(path='latexreplace')
    tar.close()

    print('Replace macro')
    print()
    texfile = 'latexreplace/draft.tex' 

    latex_replace(texfile)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : file_manager.py
# Author            : Gregory Kahn
# Email             : 
# Date              : 
# Last Modified Date: 04.12.2020
# Last Modified By  : Jiahao Yao
#
# This file is part of the VCML codebase
# Distributed under MIT license
# logging and displaying information when running

from loguru import logger
import os
import shutil
import subprocess
import sys


class FileManager(object):
    """
    def _log(self):
        logger.info('')
        logger.info('Step {0}'.format(self._get_global_step_value() - 1))
        for key, value in sorted(self._tb_logger.items(), key=lambda kv: kv[0]):
            logger.info('{0} {1:.6f}'.format(key, np.mean(value)))
        self._tb_logger.clear()

        for line in str(timeit).split('\n'):
            logger.debug(line)
        timeit.reset()
    
    Arguments:
        object {[type]} -- [description]
    
    Returns:
        [type] -- [description]
    """

    badgr_dir = os.path.abspath(__file__)[:os.path.abspath(__file__).find('src/badgr')]
    data_dir = os.path.join(badgr_dir, 'data')
    configs_dir = os.path.join(badgr_dir, 'configs')

    def __init__(self, exp_name, is_continue=False, log_fname=None, config_fname=None, add_logger=True):
        self._exp_name = exp_name
        self._exp_dir = os.path.join(self.data_dir, self._exp_name)

        if is_continue:
            assert os.path.exists(self._exp_dir),\
                'Experiment folder "{0}" does not exists, but continue = True'.format(self._exp_name)
        else:
            assert not os.path.exists(self._exp_dir),\
                'Experiment folder "{0}" exists, but continue = False'.format(self._exp_name)

        if not os.path.exists(self.git_commit_fname):
            subprocess.call('cd {0}; git log -1 > {1}'.format(FileManager.badgr_dir, self.git_commit_fname),
                            shell=True)
        if not os.path.exists(self.git_diff_fname):
            subprocess.call('cd {0}; git diff > {1}'.format(FileManager.badgr_dir, self.git_diff_fname),
                            shell=True)

        if config_fname is not None:
            shutil.copy(config_fname, os.path.join(self.exp_dir, 'config.py'))

        if add_logger:
            logger.remove()
            if log_fname:
                logger.add(os.path.join(self.exp_dir, log_fname),
                           format=self._exp_name + " {time} {level} {message}",
                           level="DEBUG")
            logger.add(sys.stdout,
                       colorize=True,
                       format="<yellow>" + self._exp_name + "</yellow> | "
                              "<green>{time:HH:mm:ss}</green> | "
                              "<blue>{level: <8}</blue> | "
                              "<magenta>{name}:{function}:{line: <5}</magenta> | "
                              "<white>{message}</white>",
                       level="DEBUG",
                       filter=lambda record: record["level"].name == "DEBUG")
            logger.add(sys.stdout,
                       colorize=True,
                       format="<yellow>" + self._exp_name + "</yellow> | "
                               "<green>{time:HH:mm:ss}</green> | "
                               "<blue>{level: <8}</blue> | "
                               "<white>{message}</white>",
                       level="INFO")

    @property
    def exp_dir(self):
        os.makedirs(self._exp_dir, exist_ok=True)
        return self._exp_dir

    ###########
    ### Git ###
    ###########

    @property
    def git_dir(self):
        git_dir = os.path.join(self.exp_dir, 'git')
        os.makedirs(git_dir, exist_ok=True)
        return git_dir

    @property
    def git_commit_fname(self):
        return os.path.join(self.git_dir, 'commit.txt')

    @property
    def git_diff_fname(self):
        return os.path.join(self.git_dir, 'diff.txt')

    ##############
    ### Models ###
    ##############

    @property
    def ckpts_dir(self):
        ckpts_dir = os.path.join(self.exp_dir, 'ckpts')
        os.makedirs(ckpts_dir, exist_ok=True)
        return ckpts_dir

    @property
    def ckpt_prefix(self):
        return os.path.join(self.ckpts_dir, 'ckpt')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : __init__.py
# Author            : Chi Han, Jiayuan Mao
# Email             : haanchi@gmail.com, maojiayuan@gmail.com
# Date              : 09.08.2019
# Last Modified Date: 09.08.2019
# Last Modified By  : Chi Han, Jiayuan Mao
#
# This file is part of the VCML codebase
# Distributed under MIT license



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : load_ckpt.py
# Author            : Chi Han, Jiayuan Mao
# Email             : haanchi@gmail.com, maojiayuan@gmail.com
# Date              : 27.07.2019
# Last Modified Date: 03.12.2019
# Last Modified By  : Chi Han
#
# This file is part of the VCML codebase
# Distributed under MIT license
#
# This file contains utility functions for loading embedding weights

import os
import torch
import numpy as np
import pandas as pd
import csv

from dataset.question_dataset.utils.misc import singular
from utility.common import make_dir


def load_embedding(filename, logger):
    '''
    with open(filename, 'r') as f:
        read_lines = f.readlines()

    data = [
        line.rstrip('\n').split(' ')
        for line in read_lines
    ]
    data = np.array(data)
    words = data[:, 0]
    weights = data[:, 1:].astype(float)
    return words, weights
    '''
    logger(f'Loading GloVe embedding from {filename}')
    table = pd.read_csv(
        filename, sep=' ', index_col=0,
        header=None, quoting=csv.QUOTE_NONE)
    return table


def align_weights(table, target, default, logger):
    logger('Aligning embedding weights from pretrained glove to model')

    def get_value(word):
        try:
            word = word.lower()
            word = singular(word)
            output = table.loc[word].values
        except Exception:
            output = default
        return output
    weight = np.stack([
        get_value(word)
        for word in target
    ])
    return weight


def download_ckpt(args, task, name, index, is_main=True):
    if task in ['CLEVR', 'GQA']:
        ckpt_name = task + '.pth'
    elif 'meronym' in name:
        ckpt_name = 'CUB_meronym.pth'
    else:
        ckpt_name = 'CUB_hypernym.pth'
    temp_dir = os.path.join(args.temp_dir, str(index))
    make_dir(temp_dir)

    ckpt_link = os.path.join(args.webpage, 'ckpt', ckpt_name)
    ckpt_file = os.path.join(temp_dir, ckpt_name)
    os.system(f'rm -f {ckpt_file}')
    if is_main:
        os.system(f'wget {ckpt_link} -P {temp_dir}')
    else:
        os.system(f'wget -q {ckpt_link} -P {temp_dir}')
    ckpt = torch.load(ckpt_file)
    os.system(f'rm {ckpt_file}')
    return ckpt



# -*- coding: utf-8 -*-
# File              : choice_utils.py
# Author            : Chi Han, Jiayuan Mao, Jiahao Yao
# Email             : haanchi@gmail.com, maojiayuan@gmail.com, jiahaoyao.math@gmail.com
# Date              : 09.08.2019
# Last Modified Date: 02.06.2020
# Last Modified By  : Jiahao Yao
#
# This file is part of the VCML codebase
# Distributed under MIT license
#
# selecting functions


import numpy as np

from .basic_utils import sum_list


def pick_one(x, requirement=None):
    if requirement:
        x = [y for y in x if requirement(y)]
    return list(x)[0]


def random_one(x, *arg, **kwarg):
    output = random_choice(x, *arg, num=1, **kwarg)
    if output is None:
        return None
    else:
        return output[0]


def group_by_value(x, fn, on_value=None):
    if on_value is None:
        def by_fn(y):
            return fn(y)
    else:
        def by_fn(y):
            return fn(on_value[y])

    all_values = set([by_fn(y) for y in x])
    output = {v: set() for v in all_values}

    for y in x:
        output[by_fn(y)].add(y)
    output = {
        k: list(v) for k, v in output.items()
    }
    return output


def random_choice(x, num, requirement=None, balance_fn=None,
                  on_value=None, replace=False, **kwarg):

    if requirement is not None:
        if on_value is None:
            x = [y for y in x if requirement(y)]
        else:
            x = [y for y in x if requirement(on_value[y])]

    if balance_fn is not None:
        group_indexes = group_by_value(x, balance_fn, on_value)
        if num > 0:
            each_num = num // len(group_indexes)
        else:
            each_num = max([len(group) for group in group_indexes.values()])
        outputs = [
            random_choice(
                group,
                num=each_num,
                replace=replace,
            )
            for group in group_indexes.values()
        ]
        return sum_list(*outputs)

    array = np.empty(len(x), 'object')
    array[:] = x
    x = array

    if len(x) == 0:
        return []
    elif num == -1:
        return x.tolist()
    elif len(x) >= num:
        return np.random.choice(x, num, replace=replace, **kwarg).tolist()
    else:
        n_round = num // x.shape[0]
        return np.concatenate(
            [x] * n_round +
            [np.random.choice(x, num-len(x)*n_round,
                              replace=replace, **kwarg)]
        ).tolist()


def random_choice_ratio(x, ratio):
    return random_choice(x, int(len(x) * ratio))

def N_splits_n(N, n):
    # use the method of wall placement
    # x1 + x2 + x3 +... + x_n = N's natural number's solutions

    all_splits_list = []
    n_splits = n - 1

    for wall_split in combinations_with_replacement(np.arange(N), n_splits):
        split = []
        for i in range(n_splits):
            if i == 0:
                split.append(wall_split[i])
            if i == n_splits-1:
                split.append(N- wall_split[i])
            else:
                split.append(wall_split[i+1] - wall_split[i])
        all_splits_list.append(split)

    return all_splits_list



import datetime
import os
import re
import time
import urllib.request as libreq
from os.path import expanduser

import certifi
import dateutil.tz
from bs4 import BeautifulSoup

import schedule


class Config():
    # listing all the topics
    topics = ['cs.LG', 'math.NA', 'math.OC', 'quant-ph']


def arxskim(topic):
    # open the corresponding html to dump the paper

    now = datetime.datetime.now(dateutil.tz.tzlocal())
    year, week, day = now.isocalendar()
    timestamp = now.strftime('%Y_%m_%d_%H_%M_%S')

    home = expanduser("~")
    folder = os.path.join(home, 'Downloads', 'arxiv_news',
                          str(year), str(week))
    os.makedirs(folder, exist_ok=True)
    text_file = open(os.path.join(folder, topic.replace(
        '.', '-') + '-' + timestamp + ".md"), "w")

    # 1000 is a dummy variaable
    with libreq.urlopen('https://arxiv.org/list/' + topic + '/pastweek?skip=0&show=1000', cafile=certifi.where()) as url:
        html = url.read()
        soup = BeautifulSoup(html, 'html.parser')

    # get the html source code
    # print(soup.prettify())

    # get the webpage
    webpage = soup.get_text()

    l_date = re.findall('\w+, \d+ \w+ \d+', webpage)
    text_file.write("### " + l_date[-1] + '-' + l_date[-5]+'\n')

    l_titles = []
    l_links = []
    l_authors = []

    for wd in webpage.split('Authors:'):
        if 'Subjects' in wd:
            l_authors.append(wd.split('Subjects')[0].replace(
                '\n', '').replace('Comments:', ',').replace('Authors:', ''))

    # delete the comments: 
    _webpage = re.sub('Comments:.*\n', '\n', webpage)

    l_titles = re.findall("Title:.+", _webpage)
    l_titles = [title.replace('Title:', '') for title in l_titles]

    l_links = re.findall('\[\d+\]\s+arXiv:\w+.\w+', _webpage)
    l_links = ["https://arxiv.org/pdf/" +
               link.split('arXiv:')[1] + '.pdf' for link in l_links]

    assert len(l_titles) == len(l_links) and len(l_titles) == len(l_authors)

    for (title, link, author) in zip(l_titles, l_links, l_authors):
        text_file.write(
            "- " + title + ', [pdf]({})'.format(link) + ', ' + author + '\n')

    text_file.close()


def job():
    config = Config()
    for topic in config.topics:
        arxskim(topic)


def main():
    schedule.every().saturday.at("12:00").do(job)

    while True:
        schedule.run_pending()
        time.sleep(1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : sys_utils.py
# Author            : Jiahao Yao
# Email             : jiahaoyao.math@gmail.com
# Date              : 04.12.2020
# Last Modified Date: 04.12.2020
# Last Modified By  : Jiahao Yao
#
# This file is part of the flow codebase
# Distributed under MIT license
#
# platform information

import os
import sys
import platform
import multiprocessing as mp
from torch.utils.collect_env import get_pretty_env_info, run, run_and_read_all


def print_sysinfo():
    # https://sebastianraschka.com/Articles/2014_multiprocessing.html#how-to-retrieve-results-in-a-particular-order
    print("\nPython version  :", platform.python_version())
    print("compiler        :", platform.python_compiler())

    print("\nsystem     :", platform.system())
    print("release    :", platform.release())
    print("machine    :", platform.machine())
    print("processor  :", platform.processor())
    print("CPU count  :", mp.cpu_count())
    print("interpreter:", platform.architecture()[0])
    print("\n\n")


_ROOT_DIR = os.path.abspath(os.path.dirname(__file__) + "/../..")


def git_available():
    try:
        run("git version")
        return True
    except:
        return False

def get_cmd_line():
    cmd_line_list = sys.argv
    cmd_line_list = ['\npython'] + cmd_line_list
    cmd_line = ' '.join(cmd_line_list)
    return cmd_line

def get_git_rev(root_dir=_ROOT_DIR, first=8):
    git_rev = run_and_read_all(run, "cd {:s} && git rev-parse HEAD".format(root_dir))
    return git_rev[:first] if git_rev else git_rev


def get_git_modifed(root_dir=_ROOT_DIR, git_dir=_ROOT_DIR):
    # Note that paths returned by git ls-files are relative to the script.
    return run_and_read_all(
        run, "cd {:s} && git ls-files {:s} -m".format(root_dir, git_dir)
    )


def get_git_untracked(root_dir=_ROOT_DIR, git_dir=_ROOT_DIR):
    # Note that paths returned by git ls-files are relative to the script.
    return run_and_read_all(
        run,
        "cd {:s} && git ls-files {:s} --exclude-standard --others".format(
            root_dir, git_dir
        ),
    )


def get_PIL_version():
    try:
        import PIL
    except ImportError:
        return "\n No Pillow is found."
    else:
        return "\nPillow ({})".format(PIL.__version__)


def collect_env_info():
    env_str = get_pretty_env_info()
    env_str += get_cmd_line()
    env_str += get_PIL_version()
    if git_available():
        env_str += "\nGit revision number: {}".format(get_git_rev())
        env_str += "\nGit Modified\n{}".format(get_git_modifed())
        # env_str += '\nGit Untrakced\n {}'.format(get_git_untracked())
    return env_str



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : interactive.py
# Author            : Jiahao Yao
# Email             : jiahaoyao.math@gmail.com
# Date              : 09.18.2020
# Last Modified Date: 09.18.2020
# Last Modified By  : Jiahao Yao
#
# This file is part of the FLOW codebase
# Distributed under MIT license

import os 
import numpy as np

def get_freer_gpu():
    os.system('nvidia-smi -q -d Memory |grep -A4 GPU|grep Free >tmp')
    memory_available = [int(x.split()[2]) for x in open('tmp', 'r').readlines()]
    return np.argmax(memory_available)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : logger.py
# Author            : Jiahao Yao
# Email             : jiahaoyao.math@gmail.com
# Date              : 03.17.2021
# Last Modified Date: 03.17.2021
# Last Modified By  : Jiahao Yao
#
# This file is part of the VCML codebase
# Distributed under MIT license
# setup logger before running experiments 

import datetime
import os
import random
import sys

import dateutil.tz
from sysflow.log.file_manager import FileManager
from sysflow.log.logger_utils import get_logger
from sysflow.utils.common_utils.file_utils import dump, load, make_dir


# set up the logger's
def exp_setup(exp_name, file_path, args):
    # file_path = os.path.realpath(__file__)
    # args: namespace

    if os.path.isdir(file_path): 
        data_dir = os.path.join(file_path, "exp")
    else: 
        data_dir = os.path.join(os.path.dirname(file_path), "exp")
    make_dir(data_dir)

    exp_dir = os.path.join(data_dir, exp_name)
    make_dir(exp_dir)

    now = datetime.datetime.now(dateutil.tz.tzlocal())
    timestamp = now.strftime("%m_%d_%Y_%H_%M_%S")
    # avoid collision
    random_int = random.randint(0, 65536)
    random_timestamp = "{}_{}".format(timestamp, random_int)
    exp_dir = os.path.join(data_dir, random_timestamp)
    make_dir(exp_dir)

    # git-info
    FileManager(exp_dir)

    args.exp_dir = exp_dir

    params = vars(args)
    # save params
    param_dir = os.path.join(exp_dir, "train_params.json")
    dump(params, param_dir)

    # logger serve as stdout
    logger = get_logger(logpath=os.path.join(exp_dir, "logs"), filepath=file_path)
    logger.info(args)
    logger.info(exp_dir)

    # sys is global vars
    cmd_line = "python " + " ".join(sys.argv)
    logger.info(cmd_line)

    return exp_dir, logger



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : colortext.py
# Author            : Chi Han, Jiayuan Mao
# Email             : haanchi@gmail.com, maojiayuan@gmail.com
# Date              : 07.08.2019
# Last Modified Date: 07.08.2019
# Last Modified By  : Chi Han, Jiayuan Mao
#
# This file is part of the VCML codebase
# Distributed under MIT license
#
# providing color configurations for texting


import copy


configs = {
    'HEADER': '\033[95m',
    'OKBLUE': '\033[94m',
    'OKGREEN': '\033[92m',
    'WARNING': '\033[93m',
    'FAIL': '\033[91m',
    'ENDC': '\033[0m',
    'BOLD': '\033[1m',
    'UNDERLINE': '\033[4m',
}


def colored(text, name):
    head = configs[name]
    tail = configs['ENDC']
    return head + text + tail


def OKGREEN(text):
    return colored(text, 'OKGREEN')


def FAIL(text):
    return colored(text, 'FAIL')


def OKBLUE(text):
    return colored(text, 'OKBLUE')


def remove_color(text):
    text = copy.deepcopy(text)
    for token in configs.values():
        text.replace(token, '')
    return text



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : logging.py
# Author            : Chi Han, Jiayuan Mao
# Email             : haanchi@gmail.com, maojiayuan@gmail.com
# Date              : 25.07.2019
# Last Modified Date: 04.12.2019
# Last Modified By  : Jiahao Yao
#
# This file is part of the VCML codebase
# Distributed under MIT license
# logging and displaying information when running


import os
import datetime
import pprint
from tqdm import tqdm

from utils.common \
    import sum_list, make_parent_dir
from utils import colortext as color


def get_levelup(logger):

    class levelup:
        def __init__(self):
            pass

        def __enter__(self):
            logger._levelup()

        def __exit__(self, _type, _value, _traceback):
            logger._leveldown()

    return levelup


class Logger:
    """
    Useage:     
    
    logger = Logger(local_dir, is_main=True)

    logger('Loading dataset')
    with logger.levelup():
        do_something

    Returns:
        [type] -- [description]
    """
    def __init__(self, log_dir, is_main=True, silent=False):
        self.logbook = []
        self.starttime = self.time
        self.level = 0
        self.levelup = get_levelup(self)
        self.log_dir = log_dir
        self.is_main = is_main
        self.silent = silent

    @property
    def time(self):
        return datetime.datetime.now()

    def time_str(self):
        return str(self.time)

    @property
    def time_message(self):
        return f'Current Time is: {self.time_str()}'

    def showtime(self):
        self.log(self.time_message)

    def _levelup(self):
        self.level += 1

    def _leveldown(self):
        self.level -= 1

    @property
    def arrow(self):
        if self.level == 0:
            return '=> '
        elif self.level == 1:
            return '-----> '
        elif self.level == 2:
            return '---------> '
        elif self.level == 3:
            return '- - - - - - -> '
        elif self.level == 4:
            return '- - - - - - - - -> '
        elif self.level == 5:
            return '-   -   -   -   -   -> '
        elif self.level == 6:
            return '-   -   -   -   -   -   -> '
        else:
            return ' ' * (self.level * 4) + '-> '

    @property
    def time_head(self):
        delta = self.time - self.starttime
        cropped = str(delta)[:10]
        return cropped

    @property
    def colored_head(self):
        time_head = self.time_head
        arrow = self.arrow
        return color.OKGREEN(time_head) + ' | ' + color.OKBLUE(arrow)

    @property
    def uncolored_head(self):
        time_head = self.time_head
        arrow = self.arrow
        return time_head + ' | ' + arrow

    @property
    def empty_time_head(self):
        time_head = self.time_head
        return ' ' * len(time_head) + ' | '

    @property
    def empty_head(self):
        arrow = self.arrow
        return self.empty_time_head + ' ' * len(arrow)

    @property
    def columns_left(self):
        width = self.get_windowsize()[1]
        return width - len(self.empty_head)

    def partition_string(self, string):
        room = self.columns_left
        ls = len(string)
        if ls < room:
            return [string]
        else:
            partitioned = [
                string[i * room: (i+1) * room]
                for i in range(ls // room + 1)
            ]
            return partitioned

    def print(self, content, pretty, resume):
        if pretty:
            string = pprint.pformat(content)
        else:
            string = str(content)

        partitioned = sum_list(
            *[self.partition_string(line)
              for line in string.split('\n')]
        )

        for i, line in enumerate(partitioned):
            self.delete_line()
            if i == 0 and not resume:
                partitioned[i] = self.uncolored_head + line
                if self.is_main:
                    print(self.colored_head + line)
            else:
                partitioned[i] = self.empty_head + line
                if self.is_main:
                    print(self.empty_head + line)
        return partitioned

    def log(self, item, pretty=False, resume=False):
        contents = self.print(item, pretty, resume)
        self.logbook.extend(contents)
        self.dump()

    def __call__(self, *arg, **kwarg):
        return self.log(*arg, **kwarg)

    def split_line(self):
        if self.is_main:
            line = '-' * self.columns_left
            print(self.empty_time_head + color.OKBLUE(line))

    def line_break(self):
        if self.is_main:
            self.delete_line()
            print(self.empty_head)

    def delete_line(self):
        if self.is_main:
            print('\r', end='')

    def state_dict(self):
        ckpt = {
            'logbook': self.logbook
        }
        return ckpt

    def load_state_dict(self, ckpt):
        self.logbook = ckpt['logbook'] + self.logbook

    def set_display(self):
        self.is_main = True

    def off_display(self):
        self.is_main = False

    @property
    def log_filename(self):
        filename = os.path.join(
            self.log_dir,
            'logging.txt'
        )
        return filename

    def dump(self):
        if not self.silent:
            make_parent_dir(self.log_filename)
            with open(self.log_filename, 'w') as f:
                for line in self.logbook:
                    f.write(color.remove_color(line) + '\n')

    def tqdm(self, iterable, *arg, **kwarg):
        output = tqdm(iterable, *arg, disable=not self.is_main,
                      leave=False, dynamic_ncols=True, **kwarg)
        return output

    @staticmethod
    def get_windowsize():
        rows, columns = os.popen('stty size', 'r').read().split()
        rows = int(rows)
        columns = int(columns)
        return rows, columns

    def get_ncols(self):
        return self.get_windowsize()[1]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : plt_utils.py
# Author            : Chi Han, Jiayuan Mao
# Email             : haanchi@gmail.com, maojiayuan@gmail.com
# Date              : 09.08.2019
# Last Modified Date: 04.10.2019
# Last Modified By  : Chi Han, Jiayuan Mao
#
# This file is part of the VCML codebase
# Distributed under MIT license
#
# printing / visualization


import matplotlib
import numpy as np


def import_plt(style="seaborn-darkgrid", backend="Qt5Agg", fontsize=10, usetex=True):
    # robust backend
    try: 
        matplotlib.use(backend)
    except: 
        try: 
            matplotlib.use('TKAgg')
        except: 
            matplotlib.use('Agg')
        
    import matplotlib.pyplot as plt

    plt.style.use(style)
    plt.rc("text", usetex=usetex)
    if usetex:
        matplotlib.rcParams["text.latex.preamble"] = [
            r"\usepackage{amssymb}",
            r"\usepackage{amsmath}",
            r"\usepackage{xcolor}",
        ]

    nice_fonts = {
    # Use LaTeX to write all text
    "text.usetex": usetex,
    "font.family": "serif",
    # Use 10pt font in plots, to match 10pt font in document
    "axes.labelsize": fontsize + 2,
    "font.size": fontsize + 2,
    # Make the legend/label fonts a little smaller
    "legend.fontsize": fontsize,
    "xtick.labelsize": fontsize,
    "ytick.labelsize": fontsize,
    }
    matplotlib.rcParams.update(nice_fonts)

    return plt




def set_size(width=345, fraction=1, subplots=(1, 1)):
    """ Set figure dimensions to avoid scaling in LaTeX.

    Parameters
    ----------
    width: float or string
            Document width in points, or string of predined document type
    fraction: float, optional
            Fraction of the width which you wish the figure to occupy
    subplots: array-like, optional
            The number of rows and columns of subplots.
    Returns
    -------
    fig_dim: tuple
            Dimensions of figure in inches
    """
    if width == "thesis":
        width_pt = 426.79135
    elif width == "beamer":
        width_pt = 307.28987
    elif width == "pnas":
        width_pt = 246.09686
    else:
        width_pt = width

    # Width of figure (in pts)
    fig_width_pt = width_pt * fraction
    # Convert from pt to inches
    inches_per_pt = 1 / 72.27

    # Golden ratio to set aesthetic figure height
    golden_ratio = (5 ** 0.5 - 1) / 2

    # Figure width in inches
    fig_width_in = fig_width_pt * inches_per_pt
    # Figure height in inches
    fig_height_in = fig_width_in * golden_ratio * (subplots[0] / subplots[1])

    return (fig_width_in, fig_height_in)


def latex(label):
    return '$\\text{' + label + '}$'

def get_axes(num, plt):
    nrows = int(np.sqrt(num))
    ncols = num // nrows + 1

    fig, ax = plt.subplots(nrows=nrows, ncols=ncols, figsize=(ncols * 6.4, nrows * 4.8))
    ax_array = ax.flatten()
    return fig, ax_array

def plt_st(nrows=1, ncols=1, sharex=False, sharey=False):
    fig, ax = plt.subplots(nrows=nrows, ncols=ncols, sharex=False, sharey=False, figsize=set_size(subplots=(nrows, ncols)))
    return fig, ax

def plt_end(filename=None, show=True, tight=True):
    if tight:
        plt.tight_layout()
    if filename:
        plt.savefig(filename+'.pdf')
        plt.savefig(filename+'.png')
    if show:
        plt.show()

plt = import_plt()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : np_utils.py
# Author            : Jiahao Yao
# Email             : jiahaoyao.math@gmail.com
# Date              : 03.31.2020
# Last Modified Date: 03.31.2020
# Last Modified By  : Jiahao Yao
#
# This file is part of the flow codebase
# Distributed under MIT license
#
# numpy tools

def matprint(mat, fmt="g"):
    # adapted from: https://gist.github.com/lbn/836313e283f5d47d2e4e
    col_maxes = [max([len(("{:"+fmt+"}").format(x)) for x in col]) for col in mat.T]
    for x in mat:
        for i, y in enumerate(x):
            print(("{:"+str(col_maxes[i])+fmt+"}").format(y), end="  ")
        print("")
        


import argparse
import os
import subprocess
import sys

from sysflow.utils.common_utils.file_utils import dump, is_empty, load

BASE_DIR = os.path.dirname(os.path.realpath(__file__))
user_config = os.path.join(BASE_DIR, "slurm_user.yaml")


def _default_item(resources, key, value):
    if key not in resources:
        resources[key] = value


def _default_append_item(resources, key, value):
    if key not in resources:
        resources[key] = value
    elif isinstance(resources[key], str):
        resources[key] = [resources[key]] + value
    elif isinstance(resources[key], list):
        resources[key] = resources[key] + value

# reference: https://docs-research-it.berkeley.edu/services/high-performance-computing/user-guide/hardware-config/
CPU_CAP = {
    'savio': 20, 
    'savio2': 24, 
    'savio3': 32
}

class Slurm:
    def __init__(self, **args_dict):
        self.config(**args_dict)

    def default_config(self, res_):
        """
        set default value if a key in res_ is not fhound
        """
        if res_ == None:
            res = {}
        else:
            res = res_
        _default_item(res, "numb_node", 1)
        _default_item(res, "job_name", "test")
        _default_item(res, "task_per_node", 32)
        _default_item(res, "time_limit", "24:00:00")
        _default_item(res, "partition", "savio3")
        _default_item(res, "account", "co_esmath")
        _default_item(res, "qos", "esmath_savio3_normal")
        _default_append_item(res, "module_list", ["gcc", "openmpi"])
        _default_item(res, "email", "jiahaoyao.math@gmail.com")
        _default_item(res, "conda_env", "qrl")

        self.limit_cpu_cap(res)
        return res

    def merge_config(self, res, res_):
        for key, value in res_.items():
            if key == "module_list":
                _default_append_item(res, key, value)
            else:
                _default_item(res, key, value)

        self.limit_cpu_cap(res)
        return res

    def limit_cpu_cap(self, res):
        if res['partition'] in CPU_CAP: 
            res['task_per_node'] = min(res['task_per_node'], CPU_CAP[res['partition']])
        return res 

    def sub_script_head(self, res):
        ret = ""
        ret += "#!/bin/bash\n"
        ret += "#SBATCH --job-name=%s \n" % res["job_name"]
        ret += "#SBATCH --account=%s \n" % res["account"]
        ret += "#SBATCH --partition=%s \n" % res["partition"]
        ret += "#SBATCH --qos=%s \n" % res["qos"]
        ret += "#SBATCH --nodes=%d\n" % res["numb_node"]
        ret += "#SBATCH --ntasks-per-node=%d\n" % res["task_per_node"]
        ret += "#SBATCH -t %s\n" % res["time_limit"]
        ret += "#SBATCH --mail-type=END,FAIL\n#SBATCH --mail-user=%s\n" % res["email"]
        ret += "\n"

        if isinstance(res["module_list"], str):
            ret += "module load %s\n" % res["module_list"]
        elif isinstance(res["module_list"], list):
            for ii in res["module_list"]:
                ret += "module load %s\n" % ii
        ret += 'eval "$(conda shell.bash hook)"\nconda activate %s\n' % res["conda_env"]
        ret += "\n"

        ret += "$@\n"
        return ret

    def config(self, **args_dict):
        if is_empty(user_config):
            self.default_config(args_dict)
            dump(args_dict, user_config)
        else:
            user_args_dict = load(user_config)
            self.merge_config(args_dict, user_args_dict)
            dump(args_dict, user_config)

        slurm_bash = self.sub_script_head(args_dict)
        dump(slurm_bash, "savio3.sh")

    def run(self, command_line, slurm=True):
        print(command_line)
        if slurm:
            subprocess.call("sbatch savio3.sh {}".format(command_line), shell=True)
        else:
            subprocess.call("{}".format(command_line), shell=True)


def config(args=None):
    parser = argparse.ArgumentParser(
        prog="slurm config",
        description="personalize configurations for the slurm.",
        argument_default=argparse.SUPPRESS,
    )
    parser.add_argument("--numb_node", type=int, help="number of nodes")
    parser.add_argument("--job_name", type=str, help="the name for the job")
    parser.add_argument("--task_per_node", type=int, help="task number for each node")
    parser.add_argument(
        "--time_limit", type=str, help="set the cap for the time, format: hour:min:sec"
    )
    parser.add_argument("--partition", type=str, help="which partition for the job")
    parser.add_argument("--account", type=str, help="which account to use")
    parser.add_argument("--qos", type=str, help="which qos in the slurm account")
    parser.add_argument(
        "--module_list",
        type=str,
        nargs="*",
        help="the list of modules to load for the slurm",
    )
    parser.add_argument(
        "--email", type=str, help="the email to notify when the jobs are finished"
    )
    parser.add_argument(
        "--conda_env", type=str, help="conda environment to activate in the scripts"
    )
    args = parser.parse_args(args)

    args_dict = vars(args)

    if "job_name" not in args_dict:
        job_name = input("Enter your job name: ")
        if job_name: 
            args_dict["job_name"] = job_name

    if "email" not in args_dict:
        email = input("Enter your email: ")
        if email: 
            args_dict["email"] = email

    if "conda_env" not in args_dict:
        conda_env = input("Enter your anaconda environment: ")
        if conda_env: 
            args_dict["conda_env"] = conda_env

    slurm = Slurm(**args_dict)


def run(args=None):
    parser = argparse.ArgumentParser(
        prog="slurm run",
        description="run a command using the slurm job submission",
        argument_default=argparse.SUPPRESS,
    )
    parser.add_argument("command", nargs="*", help="the command line to execute")

    command_line = " ".join(sys.argv[2:])

    slurm = Slurm()

    slurm.run(command_line)


def main(args=None):
    parser = argparse.ArgumentParser(
        prog="slurm", description="A program to generate slurm command lines."
    )
    parser.add_argument(
        "command",
        help="specify the sub-command to run, possible choices: " "config, run",
    )
    parser.add_argument(
        "args",
        nargs=argparse.REMAINDER,
        help="arguments to be passed to the sub-command",
    )

    args = parser.parse_args(args)

    # sepatate all sub_command to make them useable independently
    if args.command.upper() == "CONFIG":
        sub_command = config
    elif args.command.upper() == "RUN":
        sub_command = run
    else:
        return ValueError(f"unsupported sub-command: {args.command}")

    sub_command(args.args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : word_index.py
# Author            : Chi Han, Jiayuan Mao
# Email             : haanchi@gmail.com, maojiayuan@gmail.com
# Date              : 22.07.2019
# Last Modified Date: 26.11.2019
# Last Modified By  : Chi Han
#
# This file is part of the VCML codebase
# Distributed under MIT license

import numpy as np
import os
from utility.common \
    import make_dir, load, dump, contains
from copy import deepcopy

special_tokens = [
    '<NULL>', '<START>', '<END>', '<UNK>',
]


class Tools:
    """
    A kit-like class, for storing all Word_Index instances
    """
    def __init__(self, kit_config, logger):
        self.logger = logger
        self.dict = {}
        for name, config in kit_config.items():
            item = Word_Index(name, config['use_special_tokens'],
                              logger)
            setattr(self, name, item)
            self.dict[name] = item

    def save(self, path):
        local_path = os.path.join(path, 'word2index')
        make_dir(local_path)
        for item in self.dict.values():
            item.save(local_path)

    def load(self, path):
        local_path = os.path.join(path, 'word2index')
        for item in self.dict.values():
            item.load(local_path)

    def state_dict(self):
        ckpt = {name: item.state_dict()
                for name, item in self.dict.items()}
        return ckpt

    def load_state_dict(self, ckpt):
        for name, item in self.dict.items():
            item.load_state_dict(ckpt[name])

    def count(self, AinB_config):
        for name, item in self.dict.items():
            num = len(item)
            setattr(self, 'n_'+name, num)
        for A, B in AinB_config:
            AinB = A_in_B_dict(self.dict[A], self.dict[B])
            setattr(self, A+'_in_'+B, AinB)


class Word_Index:
    def __init__(self, name, use_special_tokens, logger):
        self.name = name
        self.use_special_tokens = use_special_tokens
        self.logger = logger
        self.init()

    def init(self):
        if self.use_special_tokens:
            self.records = deepcopy(special_tokens)
        else:
            self.records = []

        self.make_dict()

    def load(self, path):
        filepath = os.path.join(path, self.name+'.pkl')
        ckpt = load(filepath)
        self.load_state_dict(ckpt)
        self.logger(f'loading Word_Index: {self.name}')
        self.logger(f'from file {filepath}, length={len(self)}',
                    resume=True)

    def save(self, path):
        filepath = os.path.join(path, self.name+'.pkl')
        self.logger(f'saving Word_Index: {self.name}, '
                    f'length={len(self)}')
        dump(self.records, filepath)

    def state_dict(self):
        ckpt = {'records': self.records}
        return ckpt

    def load_state_dict(self, ckpt):
        if isinstance(ckpt, dict):
            self.records = ckpt['records']
        else:
            self.records = ckpt
        self.make_dict()

    def reset(self, items):
        self.init()
        for item in items:
            self.register(item)

    def make_dict(self):
        self.dict = {item: i for i, item in enumerate(self.records)}
        self.len = len(self.records)

    def __getitem__(self, query):
        if isinstance(query, int):
            if query < len(self):
                return self.records[query]
            else:
                if self.use_special_tokens:
                    return '<UNK>'
                else:
                    raise Exception('index out of range')

        elif isinstance(query, str):
            if query in self.dict:
                return self.dict[query]
            else:
                if self.use_special_tokens:
                    return self.dict['<UNK>']
                else:
                    raise Exception(f'{query} not in word2index {self.name}')
        elif isinstance(query, list):
            return np.array(self.records)[query].tolist()
        else:
            raise Exception(f'invalid query: {query}')

    def register(self, item, multiple=False):
        if multiple:
            for y in item:
                self.register(y)
        else:
            assert isinstance(item, str)
            if item not in self:
                self.records.append(item)
                self.dict[item] = len(self)
                self.len += 1

    def register_special(self):
        self.register(special_tokens, True)

    def register_related(self, stats):
        for key, values in stats.items():
            if key in self or contains(self, values):
                self.register(key)
                self.register(values, multiple=True)

    def __len__(self):
        return self.len

    def __contains__(self, item):
        return item in self.dict

    def __iter__(self):
        return iter(self.records)

    def indexes(self):
        return list(range(len(self)))

    def filter_out(self, forbidden):
        remained = [
            x for x in self.records if x not in forbidden
        ]
        self.reset(remained)

    def filter_most_frequent(
        self,
        word_count,
        num=-1,
    ):
        if num == -1:
            return
        else:
            remained = set()
            n = 0
            for word, count in sorted(
                word_count.items(),
                key=lambda x: x[1],
                reverse=True
            ):
                if word in self:
                    remained.add(word)
                    n += 1
                if n == num:
                    break

            self.reset(remained)

    def get_indexes(self, items):
        return [
            self.dict[item]
            for item in items
        ]


def A_in_B_dict(a, b):
    return np.array([
        b[x] if x in b
        else -1
        for x in a
    ])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : interactive.py
# Author            : Chi Han, Jiayuan Mao
# Email             : haanchi@gmail.com, maojiayuan@gmail.com
# Date              : 09.08.2019
# Last Modified Date: 09.08.2019
# Last Modified By  : Chi Han, Jiayuan Mao
#
# This file is part of the VCML codebase
# Distributed under MIT license


def yes_or_no(question):
    answer = None
    while answer is None:
        message = f'{question} [yes/no]: '
        feedback = input(message)
        if feedback == 'yes':
            answer = True
        elif feedback == 'no':
            answer = False
        else:
            print('Invalid feedback, please answer again.')

    return answer



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : logger.py
# Author            : Jiayuan Gu
# Email             : 
# Date              : 
# Last Modified Date: 04.12.2019
# Last Modified By  : Jiahao Yao
#
# This file is part of the VCML codebase
# Distributed under MIT license
# logging and displaying information when running
# https://github.com/maxjaritz/mvpnet/blob/master/common/utils/logger.py


import logging
import os
import sys


def setup_logger(name, save_dir, comment=''):
    """
    Useage:

    timestamp = time.strftime('%m-%d_%H-%M-%S')
    hostname = socket.gethostname()
    run_name = '{:s}.{:s}'.format(timestamp, hostname)

    logger = setup_logger('mvpnet', output_dir, comment='test.{:s}'.format(run_name))
    logger = setup_logger('mvpnet', output_dir, comment='train.{:s}'.format(run_name))
    logger.info('{:d} GPUs available'.format(torch.cuda.device_count()))
    logger.info(args)

    
    Arguments:
        name {[type]} -- [description]
        save_dir {[type]} -- [description]
    
    Keyword Arguments:
        comment {str} -- [description] (default: {''})
    
    Returns:
        [type] -- [description]
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s: %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    if save_dir:
        filename = 'log'
        if comment:
            filename += '.' + comment
        log_file = os.path.join(save_dir, filename + '.txt')
        fh = logging.FileHandler(log_file)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    return logger


# https://github.com/wookayin/video2gif
# http://blog.pkh.me/p/21-high-quality-gif-with-ffmpeg.html#usage

r"""
video2gif: Convert a video file into a GIF.
The GIF should be (reasonably) compressed and high-quality.
Usage:
    $ video2gif foo.mp4
    $ video2gif foo.mp4 foo.gif --scale 640
"""

import argparse
import os.path
import re
import shlex
import subprocess as sp
import sys
import tempfile


def run(cmd: str):
    sys.stderr.write(cmd)
    sys.stderr.write("\n")
    sys.stderr.flush()
    return sp.call(cmd, shell=True)


def convert(inputfile, outputfile=None, *, fps=24, scale=None):
    if not os.path.exists(inputfile):
        sys.stderr.write("File does not exist: " + inputfile + "\n")
        return 1
    if not outputfile:
        outputfile = os.path.splitext(inputfile)[0] + ".gif"

    fd, tmpfile = tempfile.mkstemp(suffix=".png", prefix="video2gif-palette")
    os.close(fd)

    try:
        palette = tmpfile
        filters = f"fps={fps}"
        if scale:
            # e.g. 50% -> iw:50, ih:50
            match = re.search(r"^(\d+\.?\d*)%$", scale)
            if match:
                percent = float(match.group(1)) / 100.0
                scale = "iw*{percent:.2f}:ih*{percent:.2f}".format(percent=percent)

            filters += ",scale={scale}:-1:flags=lanczos".format(scale=scale)

        ctx = dict(
            inputfile=shlex.quote(inputfile),
            outputfile=shlex.quote(outputfile),
            filters=filters,
            palette=palette,
        )

        cmd = 'ffmpeg -v warning -i {inputfile} -vf "{filters},palettegen" -y {palette}'.format(
            **ctx
        )
        if run(cmd) != 0:
            return 1
        cmd = 'ffmpeg -v warning -i {inputfile} -i {palette} -lavfi "{filters} [x]; [x][1:v] paletteuse" -y {outputfile}'.format(
            **ctx
        )
        if run(cmd) != 0:
            return 1

    finally:
        try:
            os.remove(tmpfile)
        except OSError:
            pass

    return 0


def parse_args():
    desc = "convert the mp4 to gif"
    parser = argparse.ArgumentParser(description=desc)
    arg_lists = []

    def add_argument_group(name):
        arg = parser.add_argument_group(name)
        arg_lists.append(arg)
        return arg

    # animation arg
    anima_arg = add_argument_group("animation")

    anima_arg.add_argument("--inputfile", "-i", type=str, help="input file")
    anima_arg.add_argument("--outputfile", "-o", type=str, help="output file")
    anima_arg.add_argument(
        "--scale",
        default=None,
        type=str,
        help="Output scale. e.g. 640, 640:360, -1:480, iw*2:ih, 50%%",
    )
    anima_arg.add_argument("--fps", default=24, type=int, help="FPS (defaults 24)")

    args = parser.parse_args()

    return args


def main():
    global args
    args = parse_args()

    convert(**vars(args))


if __name__ == "__main__":
    main()




import sys 
import argparse

def parse_args():
    desc = "change the format of the reference list to standard markdown on the web"
    parser = argparse.ArgumentParser(description=desc)
    arg_lists = []

    def add_argument_group(name):
        arg = parser.add_argument_group(name)
        arg_lists.append(arg)
        return arg

    # latex path
    arx_arg = add_argument_group("ref")
    arx_arg.add_argument("path", type=str, help="the path of the markdown file")
    arx_arg.add_argument("--sid", type=int, default=1, help="the start index of the item")
    args = parser.parse_args()

    return args


def get_doi(line_str):
    import re 
    import requests

    regex_doi = '[doi|DOI][\s\.\:]{0,2}(10\.\d{4}[\d\:\.\-\/a-zA-Z]+)'
    doi_id = re.findall(regex_doi, line_str)
    if len(doi_id) == 0:
        return False 
    # there might be some period in the end
    doi_id = doi_id[0].strip('.')

    bare_url = "http://api.crossref.org/"

    def get_json(doi):
        """
        Parameters
        ----------
            doi: str

        Returns
        -------

            found: bool
            item: dict
                Response from crossref
        """

        url = "{}works/{}"
        url = url.format(bare_url, doi)
        r = requests.get(url)
        found = False if r.status_code != 200 else True
        item = r.json()

        return found, item

    found, doi_json = get_json(doi_id)

    if found: 
        title = doi_json['message']['title'][0]
        url = doi_json['message']['URL']
        
        return '[{}]({})'.format(title, url)
    return found 



def get_arxiv(line_str):
    import re 
    import requests

    ARXIV_IDENTIFIER_FROM_2007 = r"\d{4}\.\d{4,5}(v\d+)?"
    ARXIV_IDENTIFIER_BEFORE_2007 = r"(" + ("|".join([
        "astro-ph.GA",
        "astro-ph.CO",
        "astro-ph.EP",
        "astro-ph.HE",
        "astro-ph.IM",
        "astro-ph.SR",
        "cond-math.dis-nn",
        "cond-math.mtrl-sci",
        "cond-math.mes-hall",
        "cond-math.other",
        "cond-math.quant-gas",
        "cond-math.soft",
        "cond-math.stat-mech",
        "cond-math.str-el",
        "cond-math.supr-con",
        "gr-qc",
        "hep-ex",
        "hep-lat",
        "hep-ph",
        "hep-th",
        "math-ph",
        "nlin.AO",
        "nlin.CG",
        "nlin.CD",
        "nlin.SI",
        "nlin.PS",
        "nucl-ex",
        "nucl-th",
        "physics.acc-ph",
        "physics.ao-ph",
        "physics.atom-ph",
        "physics.atm-clus",
        "physics.bio-ph",
        "physics.chem-ph",
        "physics.class-ph",
        "physics.comp-ph",
        "physics.data-an",
        "physics.flu-dyn",
        "physics.gen-ph",
        "physics.geo-ph",
        "physics.hist-ph",
        "physics.ins-det",
        "physics.med-ph",
        "physics.optics",
        "physics.ed-ph",
        "physics.soc-ph",
        "physics.plasm-ph",
        "physics.pop-ph",
        "physics.space-ph",
        "physics.quant-ph",
        "physics",
        "quant-ph",
        "math.AG",
        "math.AT",
        "math.AP",
        "math.CT",
        "math.CA",
        "math.CO",
        "math.AC",
        "math.CV",
        "math.DG",
        "math.DS",
        "math.FA",
        "math.GM",
        "math.GN",
        "math.GT",
        "math.GR",
        "math.HO",
        "math.IT",
        "math.KT",
        "math.LO",
        "math.MP",
        "math.MG",
        "math.NT",
        "math.NA",
        "math.OA",
        "math.OC",
        "math.PR",
        "math.QA",
        "math.RT",
        "math.RA",
        "math.SP",
        "math.ST",
        "math.SG",
        "cs.AI",
        "cs.CL",
        "cs.CC",
        "cs.CE",
        "cs.CG",
        "cs.GT",
        "cs.CV",
        "cs.CY",
        "cs.CR",
        "cs.DS",
        "cs.DB",
        "cs.DL",
        "cs.DM",
        "cs.DC",
        "cs.ET",
        "cs.FL",
        "cs.GL",
        "cs.GR",
        "cs.AR",
        "cs.HC",
        "cs.IR",
        "cs.IT",
        "cs.LG",
        "cs.LO",
        "cs.MS",
        "cs.MA",
        "cs.MM",
        "cs.NI",
        "cs.NE",
        "cs.NA",
        "cs.OS",
        "cs.OH",
        "cs.PF",
        "cs.PL",
        "cs.RO",
        "cs.SI",
        "cs.SE",
        "cs.SD",
        "cs.SC",
        "cs.SY",
        "q-bio.BM",
        "q-bio.CB",
        "q-bio.GN",
        "q-bio.MN",
        "q-bio.NC",
        "q-bio.OT",
        "q-bio.PE",
        "q-bio.QM",
        "q-bio.SC",
        "q-bio.TO",
        "q-fin.CP",
        "q-fin.EC",
        "q-fin.GN",
        "q-fin.MF",
        "q-fin.PM",
        "q-fin.PR",
        "q-fin.RM",
        "q-fin.ST",
        "q-fin.TR",
        "stat.AP",
        "stat.CO",
        "stat.ML",
        "stat.ME",
        "stat.OT",
        "stat.TH"])) + r")/\d+"
    # Regex is fully enclosed in a group for findall to match it all
    regex_arxiv = "((arxiv:)?((" + ARXIV_IDENTIFIER_FROM_2007 + ")|(" + ARXIV_IDENTIFIER_BEFORE_2007 + ")))"

    # Base arXiv URL used as id sometimes
    ARXIV_URL = "http://arxiv.org/abs/{arxiv_id}"
    # Eprint URL used to download sources
    ARXIV_EPRINT_URL = "http://arxiv.org/e-print/{arxiv_id}"


    arxiv_id = re.findall(regex_arxiv, line_str)

    # print(arxiv_id)
    if len(arxiv_id) == 0:
        return False 
    # there might be some period in the end
    arxiv_id = arxiv_id[0][0]


    from xml.etree import ElementTree
    import sys
    import re
    import os

    if sys.version_info < (2, 6):
        raise Exception("Python 2.6 or higher required")

    # Python 2 compatibility code
    PY2 = sys.version_info[0] == 2
    if not PY2:
        from urllib.parse import urlencode
        from urllib.request import urlopen
        from urllib.error import HTTPError
        print_bytes = lambda s: sys.stdout.buffer.write(s)
    else:
        from urllib import urlencode
        from urllib2 import HTTPError, urlopen
        print_bytes = lambda s: sys.stdout.write(s)



    def arxiv_request(ids):
        """Sends a request to the arxiv API."""
        q = urlencode([
            ("id_list", ",".join(ids))
            ])
        xml = urlopen("http://export.arxiv.org/api/query?" + q)
        # xml.read() returns bytes, but ElementTree.fromstring decodes
        # to unicode when needed (python2) or string (python3)
        return ElementTree.fromstring(xml.read())
    
    ATOM = '{http://www.w3.org/2005/Atom}'

    xml = arxiv_request([arxiv_id])

    entries = xml.findall(ATOM + "entry")

    title = entries[0].find(ATOM + 'title').text.strip()

    url = 'https://arxiv.org/abs/%s' % arxiv_id

    return '[{}]({})'.format(title, url)

def ref2md(filein, sid):
    assert filein.endswith('.md') 
    fileout = filein.replace(".md", "_out.md")
    
    sys.stdout = open(fileout, "w")

    count = sid
    with open(filein, 'r') as f: 
        for line in f: 
            # skip the empty ones
            if line.strip() == '':
                continue
            
            line_doi = get_doi(line)
            line_arxiv = get_arxiv(line)

            if line_doi:
                print('{}. {}'.format(count, line_doi))
            elif line_arxiv:
                print('{}. {}'.format(count, line_arxiv))
            else:
                print(line)
            print()

            count += 1

    sys.stdout.close()


def main():
    global args
    args = parse_args()

    ref2md(args.path, args.sid)



if __name__ == "__main__":
    # usage:
    # 1. download the file into file.md

    md_file = """
1. A. Kulesza, B. Taskar, Determinantal point processes for machine learning, Found. Trends Mach. Learn. 5 (2012) 123–286. doi:10.1561/2200000044.

2. J. Fei, C.-N. Yeh, E. Gull, Nevanlinna Analytical Continuation, (2020) 1–6. http://arxiv.org/abs/2010.04572.

3. J. Nigam, S. Pozdnyakov, M. Ceriotti, Recursive evaluation and iterative contraction of $N$-body equivariant features, 121101 (2020). doi:10.1063/5.0021116.

4. S. Becker, M. Embree, J. Wittsten, M. Zworski, Spectral characterization of magic angles in twisted bilayer graphene, (2020) 1–14. http://arxiv.org/abs/2010.05279.

5. R.A. Lang, I.G. Ryabinkin, A.F. Izmaylov, Unitary transformation of the electronic Hamiltonian with an exact quadratic truncation of the Baker-Campbell-Hausdorff expansion, (2020) 1–38. doi:10.1021/acs.jctc.0c00170.

6. P. Jakobsen, F. Jensen, Representing Exact Electron Densities by a Single Slater Determinant in Finite Basis Sets, (2020). doi:10.1021/acs.jctc.0c01029.

7. D. Chaykin, C. Jansson, F. Keil, M. Lange, K.T. Ohlhus, S.M. Rump, Rigorous Lower Bounds for the Ground State Energy of Molecules by Employing Necessary N-Representability Conditions, J. Chem. Theory Comput. (2020). doi:10.1021/acs.jctc.0c00497.

8. P. V. Sriluckshmy, M. Nusspickel, E. Fertitta, G.H. Booth, Fully Algebraic and Self-consistent Effective Dynamics in a Static Quantum Embedding, (2020). http://arxiv.org/abs/2012.05837.

9. H.K. Tran, H.-Z. Ye, T. Van Voorhis, Bootstrap embedding with an unrestricted mean-field bath, J. Chem. Phys. 153 (2020) 214101. doi:10.1063/5.0029092.

10. D. Luo, G. Carleo, B.K. Clark, J. Stokes, Gauge equivariant neural networks for quantum lattice gauge theories, (2020). http://arxiv.org/abs/2012.05232.

11. Q. Gu, L. Zhang, J. Feng, Neural network representation of electronic structure from ab initio molecular dynamics, (2020) 1–8. http://arxiv.org/abs/2011.13774.

12. B. Leimkuhler, M. Sachs, Efficient Numerical Algorithms for the Generalized Langevin Equation, (2020) 1–33. http://arxiv.org/abs/2012.04245.

13. G.Q. Ai, et al Accurately computing electronic properties of materials using eigenenergies, (2020). https://arxiv.org/abs/2012.00921

14. Y. Chen, et al, Spectral signatures of many-body localization with interacting photons, (n.d.). https://arxiv.org/pdf/1709.07108.pdf

15. A. Hashim, R.K. Naik, A. Morvan, J.-L. Ville, B. Mitchell, J.M. Kreikebaum, M. Davis, E. Smith, C. Iancu, K.P. O’Brien, I. Hincks, J.J. Wallman, J. Emerson, I. Siddiqi, Randomized compiling for scalable quantum computing on a noisy superconducting quantum processor, (2020). http://arxiv.org/abs/2010.00215.

16. A. Nahum, S. Vijay, J. Haah, Operator Spreading in Random Unitary Circuits, Phys. Rev. X. 8 (2018) 021014. doi:10.1103/PhysRevX.8.021014.

17. C.W. Von Keyserlingk, T. Rakovszky, F. Pollmann, S.L. Sondhi, Operator Hydrodynamics, OTOCs, and Entanglement Growth in Systems without Conservation Laws, Phys. Rev. X. 8 (2018) 21013. doi:10.1103/PhysRevX.8.021013.

18. F. Arute, et al Observation of separated dynamics of charge and spin in the Fermi-Hubbard model, (2020). http://arxiv.org/abs/2010.07965.

20. S. Lu, M.C. Bañuls, J.I. Cirac, Algorithms for quantum simulation at finite energies, ArXiv. (2020) 1–18.

21. I.D. Kivlichan, N. Wiebe, R. Babbush, A. Aspuru-Guzik, Bounding the costs of quantum simulation of many-body physics in real space, J. Phys. A Math. Theor. 50 (2017) 305301. doi:10.1088/1751-8121/aa77b8.

22. M. Kolodrubetz, D. Sels, P. Mehta, A. Polkovnikov, Geometry and non-adiabatic response in quantum and classical systems, Phys. Rep. 697 (2017) 1–87. doi:10.1016/j.physrep.2017.07.001.

23. G. Verdon, J. Marks, S. Nanda, S. Leichenauer, J. Hidary, Quantum Hamiltonian-Based Models & the Variational Quantum Thermalizer Algorithm, (2019) 1–21.

24. X. Yuan, S. Endo, Q. Zhao, Y. Li, S.C. Benjamin, Theory of variational quantum simulation, Quantum. 3 (2019) 191. doi:10.22331/q-2019-10-07-191.

25. T. Giurgica-Tiron, I. Kerenidis, F. Labib, A. Prakash, W. Zeng, Low depth algorithms for quantum amplitude estimation, (2020) 1–27. http://arxiv.org/abs/2012.03348.

26. K. Bharti, T. Haug, Iterative Quantum Assisted Eigensolver, (2020) 1–10. http://arxiv.org/abs/2010.05638.

27. A. Roggero, C. Gu, A. Baroni, T. Papenbrock, Preparation of excited states on a quantum computer, (2020) 1–25. http://arxiv.org/abs/2009.13485.

28. S. Chakraborty, L. Novo, J. Roland, Finding a marked node on any graph via continuous-time quantum walks, Phys. Rev. A. 102 (2020). doi:10.1103/PhysRevA.102.022227.

29. A. Kardashin, A. Uvarov, D. Yudin, J. Biamonte, Certified variational quantum algorithms for eigenstate preparation, ArXiv. (2020) 1–9. doi:10.1103/PhysRevA.102.052610.

30. C. Lin, D. Sels, Y. Ma, Y. Wang, Stochastic optimal control formalism for an open quantum system, Phys. Rev. A. 102 (2020) 52605. doi:10.1103/PhysRevA.102.052605.

31. J. Helsen, I. Roth, E. Onorati, A.H. Werner, J. Eisert, A general framework for randomized benchmarking, (2020) 1–60. http://arxiv.org/abs/2010.07974.

32. D. Grinko, J. Gacon, C. Zoufal, S. Woerner, Iterative quantum amplitude estimation, ArXiv. (2019) 1–13.

33. B. Collins, P. Śniady, Integration with respect to the Haar measure on unitary, orthogonal and symplectic group, Commun. Math. Phys. 264 (2006) 773–795. doi:10.1007/s00220-006-1554-3.

34. Simulating Large Quantum Circuits on a Small Quantum Computer, https://arxiv.org/abs/1904.00102
    """

    # testing 
    fname = 'ref.md'
    sys.stdout = open(fname, "w")
    print(md_file)
    sys.stdout.close()
    

    ref2md(fname, 1)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : knowledge_utils.py
# Author            : Chi Han, Jiayuan Mao
# Email             : haanchi@gmail.com, maojiayuan@gmail.com
# Date              : 09.08.2019
# Last Modified Date: 08.10.2019
# Last Modified By  : Chi Han, Jiayuan Mao
#
# This file is part of the VCML codebase
# Distributed under MIT license
#
# utilities about knowledge


from utility.common import belongs_to, union


def all_concepts(stats, requirement_on_key=None):
    keys = list(stats.keys())
    if requirement_on_key is not None:
        keys = list(filter(requirement_on_key, keys))

    elements_gather = [e for k in keys for e in stats[k]]

    output = list(set(elements_gather + keys))
    return output


def all_elements(stats, requirement_on_key=None):
    keys = list(stats.keys())
    if requirement_on_key is not None:
        keys = list(filter(requirement_on_key, keys))

    elements_gather = [e for k in keys for e in stats[k]]

    output = list(set(elements_gather))
    return output


def all_secondary_elements(
        stats, requirement_on_key=None, with_keys=False):
    keys = list(stats.keys())
    if requirement_on_key is not None:
        keys = list(filter(requirement_on_key, keys))

    elements_gather = [e for k in keys
                       for group in stats[k].values()
                       for e in group
                       ]

    output = list(set(elements_gather))
    if with_keys:
        output = union(output, keys)

    return output


def all_in_hierarchy(hierarchy):
    output = set()

    def trace_down(forest):
        for sub_root, sub_forest in forest.items():
            output.add(sub_root)
            if sub_forest is not None and len(sub_forest) > 0:
                trace_down(sub_forest)
    trace_down(hierarchy)
    output = list(output)
    return output


def filter_stats(stats, all_concepts):
    output = {
        key: [
            v for v in values
            if v in all_concepts
        ]
        for key, values in stats.items()
    }
    return output


def is_synonym(concept1, concept2, synonym_stats):
    output = belongs_to(synonym_stats, concept1) == \
        belongs_to(synonym_stats, concept2) and \
        belongs_to(synonym_stats, concept1) is not None
    return output


def is_instance_of(concept1, concept2,
                   isinstanceof_stats, synonym_stats):
    synset = get_synset(concept1, synonym_stats)
    output = False
    for concept in synset:
        if concept in isinstanceof_stats.get(concept2, []):
            output = True
    return output


def is_hypernym(concept2, concept1, hypernym_stats, synonym_stats):
    synset1 = get_synset(concept1, synonym_stats)
    synset2 = get_synset(concept2, synonym_stats)
    output = False
    for syn1 in synset1:
        for syn2 in synset2:
            if syn1 in hypernym_stats.get(syn2, []):
                output = True
    return output


def is_samekind(concept1, concept2,
                isinstanceof_stats, synonym_stats):
    synset1 = get_synset(concept1, synonym_stats)
    synset2 = get_synset(concept2, synonym_stats)
    for syn1 in synset1:
        if belongs_to(isinstanceof_stats, syn1) is not None:
            for syn2 in synset2:
                if belongs_to(isinstanceof_stats, syn2) is not None:
                    return belongs_to(isinstanceof_stats, syn1) == \
                        belongs_to(isinstanceof_stats, syn2)
    return None


def is_composed_of(concept2, concept1, meronym_stats):
    if concept2 in meronym_stats:
        if concept1 in meronym_stats[concept2]['true']:
            output = True
        elif concept1 in meronym_stats[concept2]['false']:
            output = False
        else:
            output = None
    else:
        output = None
    return output


def get_synset(concept, synonym_stats):
    if synonym_stats is not None and \
            belongs_to(synonym_stats, concept) is not None:
        synset = set(synonym_stats[belongs_to(
            synonym_stats, concept)])
    else:
        synset = set()
    synset.add(concept)
    return synset



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : common.py
# Author            : Chi Han, Jiayuan Mao, Jiahao Yao
# Email             : haanchi@gmail.com, maojiayuan@gmail.com, jiahaoyao.math@gmail.com
# Date              : 17.07.2019
# Last Modified Date: 03.31.2020
# Last Modified By  : Jiahao Yao
#
# This file is part of the VCML codebase
# Distributed under MIT license
#
# All common utililty function used in this code


from .common_utils.basic_utils import *
from .common_utils.choice_utils import *
from .common_utils.file_utils import *
from .common_utils.knowledge_utils import *
from .common_utils.math_utils import *
from .common_utils.np_utils import *
from .common_utils.plt_utils import *
from .common_utils.tensor_utils import *
from .common_utils.interactive import *



import argparse
import datetime
import re
import urllib.request as libreq

import certifi
import dateutil.tz
from bs4 import BeautifulSoup


def parse_args():
    desc = "download for an arxiv paper"
    parser = argparse.ArgumentParser(description=desc)
    arg_lists = []

    def add_argument_group(name):
        arg = parser.add_argument_group(name)
        arg_lists.append(arg)
        return arg

    # qunatum control
    arx_arg = add_argument_group('arxiv')
    arx_arg.add_argument('arxnum', type=str, help='arxiv number of papers')
    args = parser.parse_args()

    return args


def main():
    global args
    args = parse_args()


    import urllib.request
    import os 

    url = 'https://arxiv.org/e-print/' + args.arxnum
    fname = '{}.tar.gz'.format(args.arxnum)
    urllib.request.urlretrieve(url, fname)

    import tarfile

    tar = tarfile.open(fname, "r:gz")
    tar.extractall(path=args.arxnum)
    tar.close()
    os.remove(fname)

    url = 'https://arxiv.org/pdf/' + args.arxnum + '.pdf'
    fname = '{}/{}.pdf'.format(args.arxnum, args.arxnum)
    urllib.request.urlretrieve(url, fname)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : file_manager.py
# Author            : Gregory Kahn
# Email             :
# Date              :
# Last Modified Date: 03.17.2021
# Last Modified By  : Jiahao Yao
#
# This file is part of the VCML codebase
# Distributed under MIT license
# logging and displaying information when running

# from loguru import logger
import os
import subprocess
import sys


class FileManager(object):
    """
    def _log(self):
        logger.info('')
        logger.info('Step {0}'.format(self._get_global_step_value() - 1))
        for key, value in sorted(self._tb_logger.items(), key=lambda kv: kv[0]):
            logger.info('{0} {1:.6f}'.format(key, np.mean(value)))
        self._tb_logger.clear()

        for line in str(timeit).split('\n'):
            logger.debug(line)
        timeit.reset()

    Arguments:
        object {[type]} -- [description]

    Returns:
        [type] -- [description]
    """

    badgr_dir = "."

    def __init__(self, exp_dir):
        self._exp_dir = exp_dir

        if not os.path.exists(self.cmd_fname):
            with open(self.cmd_fname, "w") as f:
                f.write(self.get_cmd_line())

        if not os.path.exists(self.git_commit_fname):
            subprocess.call(
                "cd {0}; git log --graph --pretty=format:'%C(bold red)%h%Creset -%C(yellow)%d%Creset %s %Cgreen(%cr) %C(bold blue)<%an>%Creset' --decorate  --abbrev-commit > {1}".format(
                    exp_dir, self.git_commit_fname
                ),
                shell=True,
            )
        if not os.path.exists(self.git_diff_fname):
            subprocess.call(
                "cd {0}; git diff > {1}".format(exp_dir, self.git_diff_fname),
                shell=True,
            )

        if not os.path.exists(self.git_status_fname):
            subprocess.call(
                "cd {0}; git status > {1}".format(exp_dir, self.git_status_fname),
                shell=True,
            )

    @property
    def exp_dir(self):
        os.makedirs(self._exp_dir, exist_ok=True)
        return self._exp_dir

    @property
    def cmd_dir(self):
        cmd_dir = os.path.join(self.exp_dir, "script")
        os.makedirs(cmd_dir, exist_ok=True)
        return cmd_dir

    @property
    def cmd_fname(self):
        return os.path.join(self.cmd_dir, "script.sh")

    def get_cmd_line(self):
        cmd_line_list = sys.argv
        cmd_line_list = ["\npython"] + cmd_line_list
        cmd_line = " ".join(cmd_line_list)
        return cmd_line

    ###########
    ### Git ###
    ###########

    @property
    def git_dir(self):
        git_dir = os.path.join(self.exp_dir, "git")
        os.makedirs(git_dir, exist_ok=True)
        return git_dir

    @property
    def git_commit_fname(self):
        return os.path.join(self.git_dir, "commit.txt")

    @property
    def git_diff_fname(self):
        return os.path.join(self.git_dir, "diff.txt")

    @property
    def git_status_fname(self):
        return os.path.join(self.git_dir, "status.txt")

    ##############
    ### Models ###
    ##############

    @property
    def ckpts_dir(self):
        ckpts_dir = os.path.join(self.exp_dir, "ckpts")
        os.makedirs(ckpts_dir, exist_ok=True)
        return ckpts_dir

    @property
    def ckpt_prefix(self):
        return os.path.join(self.ckpts_dir, "ckpt")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : logger_utils.py
# Author            : Jiahao Yao
# Email             : jiahaoyao.math@gmail.com
# Date              : 03.17.2021
# Last Modified Date: 03.17.2021
# Last Modified By  : Jiahao Yao
#
# This file is part of the VCML codebase
# Distributed under MIT license
# logging and displaying information when running
import os 
import logging


def get_logger(
    logpath, filepath, package_files=[], displaying=True, saving=True, debug=False
):
    logger = logging.getLogger()
    if debug:
        level = logging.DEBUG
    else:
        level = logging.INFO
    logger.setLevel(level)
    if saving:
        info_file_handler = logging.FileHandler(logpath, mode="a")
        info_file_handler.setLevel(level)
        logger.addHandler(info_file_handler)
    if displaying:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)
        logger.addHandler(console_handler)
    logger.info(filepath)
    if os.path.isfile(filepath): 
        with open(filepath, "r") as f:
            logger.info(f.read())

    for f in package_files:
        logger.info(f)
        with open(f, "r") as package_f:
            logger.info(package_f.read())

    return logger


def setup_logger(name, log_file, level=logging.INFO):
    """Function setup as many loggers as you want"""
    # usage: logger = setup_logger("train", os.path.join(exp_dir, "train.log"))
    formatter = logging.Formatter("%(asctime)s %(message)s")

    handler = logging.FileHandler(log_file)
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)

    return logger



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : __init__.py
# Author            : Jiahao Yao
# Email             : jiahaoyao.math@gmail.com
# Date              : 02.04.2021
# Last Modified Date: 02.04.2021
# Last Modified By  : Jiahao Yao
#
# This file is part of the VCML codebase
# Distributed under MIT license



import argparse
import os
import re

from PIL import Image


def parse_args():
    desc = "convert the png's to gif"
    parser = argparse.ArgumentParser(description=desc)
    arg_lists = []

    def add_argument_group(name):
        arg = parser.add_argument_group(name)
        arg_lists.append(arg)
        return arg

    # animation arg
    anima_arg = add_argument_group("animation")

    anima_arg.add_argument("--input", "-i", type=str, help="input file")
    anima_arg.add_argument("--output", "-o", type=str, help="output file")

    args = parser.parse_args()

    return args


def png2gif(in_folder, out_file):
    """convert the figures in the in_folder folder to out_file in the gif format

    Args:
        in_folder (str): figure folder, the figures are in the format abc0001.png, abc0002.png, ....
        out_file (str): output file name
    """

    in_files = os.listdir(in_folder)
    in_png = filter(lambda x: x.endswith(".png"), in_files)
    in_png = map(lambda x: os.path.join(in_folder, x), in_png)
    in_png = sorted(in_png, key=lambda a: int(re.findall("\d+", a)[-1]))

    images = []
    for n, a in enumerate(in_png):
        exec("a" + str(n) + ' = Image.open("' + '{}")'.format(a))
        images.append(eval("a" + str(n)))

    images[0].save(
        out_file if out_file else "out.gif",
        save_all=True,
        append_images=images[1:],
        duration=50,
        loop=0,
    )


def main():
    global args
    args = parse_args()

    png2gif(args.input, args.output)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : tb_logger.py
# Author            : Jiahao Yao
# Email             : jiahaoyao.math@gmail.com
# Date              : 03.31.2020
# Last Modified Date: 03.31.2020
# Last Modified By  : Jiahao Yao
#
# This file is part of the VCML codebase
# Distributed under MIT license

# Tensorflow logger, code adopted from
# https://github.com/yunjey/pytorch-tutorial/blob/master/tutorials/04-utils/tensorboard/logger.py
# https://github.com/Glaciohound/VCML/blob/master/reason/utils/logger.py

import tensorflow as tf
import numpy as np
import scipy.misc 
try:
    from StringIO import StringIO  # Python 2.7
except ImportError:
    from io import BytesIO         # Python 3.x


class Logger(object):
    """use the tensorboard backend to visualize non-tenosrflow tensor
    
    Minimal Example: 
    logger = Logger('./logs')

    # Start training
    for step in range(total_step):
        
        #logging 
        if (step+1) % 100 == 0:
            print ('Step [{}/{}], Loss: {:.4f}, Acc: {:.2f}' 
                .format(step+1, total_step, loss.item(), accuracy.item()))

            # ================================================================== #
            #                        Tensorboard Logging                         #
            # ================================================================== #

            # 1. Log scalar values (scalar summary)
            info = { 'loss': loss.item(), 'accuracy': accuracy.item() }

            for tag, value in info.items():
                logger.scalar_summary(tag, value, step+1)

            # 2. Log values and gradients of the parameters (histogram summary)
            for tag, value in model.named_parameters():
                tag = tag.replace('.', '/')
                logger.histo_summary(tag, value.data.cpu().numpy(), step+1)
                logger.histo_summary(tag+'/grad', value.grad.data.cpu().numpy(), step+1)

            # 3. Log training images (image summary)
            info = { 'images': images.view(-1, 28, 28)[:10].cpu().numpy() }

            for tag, images in info.items():
                logger.image_summary(tag, images, step+1)
    """

    def __init__(self, log_dir):
        """Create a summary writer logging to log_dir."""
        self.writer = tf.summary.FileWriter(log_dir)

    def scalar_summary(self, tag, value, step):
        """Log a scalar variable."""
        summary = tf.Summary(value=[tf.Summary.Value(tag=tag, simple_value=value)])
        self.writer.add_summary(summary, step)

    def image_summary(self, tag, images, step):
        """Log a list of images."""

        img_summaries = []
        for i, img in enumerate(images):
            # Write the image to a string
            try:
                s = StringIO()
            except:
                s = BytesIO()
            scipy.misc.toimage(img).save(s, format="png")

            # Create an Image object
            img_sum = tf.Summary.Image(encoded_image_string=s.getvalue(),
                                         height=img.shape[0],
                                         width=img.shape[1])
            # Create a Summary value
            img_summaries.append(tf.Summary.Value(tag='%s/%d' % (tag, i), image=img_sum))

        # Create and write Summary
        summary = tf.Summary(value=img_summaries)
        self.writer.add_summary(summary, step)
        
    def histo_summary(self, tag, values, step, bins=1000):
        """Log a histogram of the tensor of values."""

        # Create a histogram using numpy
        counts, bin_edges = np.histogram(values, bins=bins)

        # Fill the fields of the histogram proto
        hist = tf.HistogramProto()
        hist.min = float(np.min(values))
        hist.max = float(np.max(values))
        hist.num = int(np.prod(values.shape))
        hist.sum = float(np.sum(values))
        hist.sum_squares = float(np.sum(values**2))

        # Drop the start of the first bin
        bin_edges = bin_edges[1:]

        # Add bin edges and counts
        for edge in bin_edges:
            hist.bucket_limit.append(edge)
        for c in counts:
            hist.bucket.append(c)

        # Create and write Summary
        summary = tf.Summary(value=[tf.Summary.Value(tag=tag, histo=hist)])
        self.writer.add_summary(summary, step)
        self.writer.flush()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : tensor_utils.py
# Author            : Chi Han, Jiayuan Mao
# Email             : haanchi@gmail.com, maojiayuan@gmail.com
# Date              : 09.08.2019
# Last Modified Date: 02.10.2019
# Last Modified By  : Chi Han, Jiayuan Mao
#
# This file is part of the VCML codebase
# Distributed under MIT license
#
# pytorch/numpy utilities


import numpy as np
import torch
import torch.nn.functional as F
from pprint import pprint


def to_tensor(x):
    if isinstance(x, torch.Tensor):
        return x
    elif isinstance(x, list):
        if isinstance(x[0], float):
            return torch.Tensor(x)
        elif isinstance(x[0], int):
            return torch.LongTensor(x)
        else:
            return x
    elif isinstance(x, np.ndarray):
        if x.dtype.char in ['d', 'f']:
            return torch.Tensor(x)
        elif x.dtype.char in ['l', 'b']:
            return torch.LongTensor(x)
        else:
            raise Exception('not convertable')
    elif isinstance(x, int) or isinstance(x, float) \
            or np.isscalar(x):
        return torch.tensor(x)
    else:
        raise Exception('not convertable')


def detach(tensor):
    if isinstance(tensor, torch.Tensor):
        if tensor.device.type == 'cuda':
            tensor = tensor.cpu()
        if tensor.requires_grad:
            tensor = tensor.detach()
        tensor = tensor.numpy()

    return tensor


def matmul(*mats):
    output = mats[0]
    for x in mats[1:]:
        if isinstance(output, torch.Tensor):
            output = torch.matmul(output, x)
        else:
            output = np.matmul(output, x)
    return output


def to_numpy(x):
    if isinstance(x, np.ndarray):
        return x
    elif isinstance(x, list):
        return np.array(x)
    elif isinstance(x, torch.Tensor):
        return x.cpu().detach().numpy()
    elif isinstance(x, torch.autograd.Variable):
        return x.data.cpu().numpy()


def to_normalized(x):
    if isinstance(x, torch.Tensor):
        return F.normalize(x, dim=-1)
    elif isinstance(x, np.ndarray):
        return to_normalized(torch.Tensor(x)).numpy()
    else:
        raise Exception('unsupported type: %s' % str(type(x)))


def init_seed(n=-1, index=-1):
    if n != -1:
        if index != -1:
            seed = n + index
        else:
            seed = n
        torch.manual_seed(seed)
        np.random.seed(seed)


def is_cuda(x):
    return x.device.type == 'cuda'


def valid_part(x, assert_finite=False):
    output = torch.isnan(x).bitwise_not() * (x.abs() != float('inf'))
    if assert_finite:
        output = output * (x.abs() < 1e10)
    return output


def is_valid_value(x, assert_finite):
    if not valid_part(x, assert_finite).all():
        return False
    else:
        return True


def assert_valid_value(*values, assert_finite=False):
    for i, x in enumerate(values):
        if not is_valid_value(x, assert_finite):
            pprint(values)
            print('Invalid tensor is', i)
            raise Exception('invalid value')


def index_by_list(tensor, indexes):
    if isinstance(tensor, torch.Tensor) or \
            isinstance(tensor, np.ndarray):
        return tensor[indexes]
    elif isinstance(tensor, list) or \
            isinstance(tensor, tuple):
        return [tensor[ind] for ind in indexes]
    else:
        raise Exception()


#  tools for visualization
import matplotlib
from matplotlib.lines import Line2D
def plot_grad_flow(named_parameters):
    '''Plots the gradients flowing through different layers in the net during training.
    Can be used for checking for possible gradient vanishing / exploding problems.

    Usage: Plug this function in Trainer class after loss.backwards() as
    "plot_grad_flow(self.model.named_parameters())" to visualize the gradient flow'''
    ave_grads = []
    max_grads= []
    layers = []
    for n, p in named_parameters:
        if(p.requires_grad) and ("bias" not in n):
            layers.append(n)
            ave_grads.append(p.grad.abs().mean())
            max_grads.append(p.grad.abs().max())
    plt.bar(np.arange(len(max_grads)), max_grads, alpha=0.1, lw=1, color="c")
    plt.bar(np.arange(len(max_grads)), ave_grads, alpha=0.1, lw=1, color="b")
    plt.hlines(0, -0.5, len(ave_grads)+0.5, lw=2, color="k" )
    plt.xticks(range(0,len(ave_grads), 1), layers)
    plt.xlim(left=-0.5, right=len(ave_grads)-0.5)
    plt.ylim(bottom = -0.001, top=0.10) # zoom in on the lower gradient regions
    plt.xlabel("Layers")
    plt.ylabel("average gradient")
    plt.title("Gradient flow")
    plt.grid(True)
    plt.legend([Line2D([0], [0], color="c", lw=4),
                Line2D([0], [0], color="b", lw=4),
                Line2D([0], [0], color="k", lw=4)], ['max-gradient', 'mean-gradient', 'zero-gradient'])
    plt.tight_layout()
    plt.show()


device = None

def init_gpu(use_gpu=True, gpu_id=0):
    global device
    if torch.cuda.is_available() and use_gpu:
        device = torch.device("cuda:" + str(gpu_id))
        print("Using GPU id {}".format(gpu_id))
    else:
        device = torch.device("cpu")
        print("GPU not detected. Defaulting to CPU.")


def set_device(gpu_id):
    torch.cuda.set_device(gpu_id)

class Ipdb(nn.Module):
    def __init__(self):
        super().__init__()
    def forward(self, x):
        import ipdb; ipdb.set_trace()
        return x
